#!/usr/bin/python
# -*- coding: utf-8 -*-
# #############################################################################
# Copyright (c) 2008, Kevin Horton
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# *
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The name of Kevin Horton may not be used to endorse or promote products
#       derived from this software without specific prior written permission.
# *
# THIS SOFTWARE IS PROVIDED BY KEVIN HORTON ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL KEVIN HORTON BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# #############################################################################
#
# version 0.11, 25 Apr 2008
#
# Version History:
# vers     date      Notes
# 0.10   17 Mar 08   Initial version.
#
# 0.11   25 Apr 08   Added ssec module (mostly empty yet)
# #############################################################################
#
# Distribution Notes
#
# HTML docs are created with epydoc, via
# "epydoc --no-private -n AeroCalc -u 'http://www.kilohotel.com/python/aerocalc/' aerocalc"
#
# Generate distribution - "python setup.py sdist"
#
# #############################################################################

"""Various aeronautical engineering calculations

This package contains the following modules:

airspeed  -       airspeed conversions and calculations.  Provides interactive
                  mode when run directly, e.g. 'python airspeed.py'
default_units -   defines default units to be used by all modules.  May be
                  overridden by a user units file.
ssec -            static source error correction calculations
std_atm -         standard atmosphere parametres and calculations
unit_conversion - convert various aeronautical parametres between commonly
                  used units
val_input -       validates user input when in interactive mode

Author: Kevin Horton
E-mail: kevin01 -at- kilohotel.com
"""

VERSION = "0.11"




#!/usr/bin/python
# -*- coding: utf-8 -*-

# #############################################################################
# Copyright (c) 2008, Kevin Horton
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# *
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The name of Kevin Horton may not be used to endorse or promote products
#       derived from this software without specific prior written permission.
# *
# THIS SOFTWARE IS PROVIDED BY KEVIN HORTON ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL KEVIN HORTON BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# #############################################################################
#
# version 0.10, 17 May 2007
#
# Version History:
# vers     date      Notes
# 0.10   17 May 07   Initial version.
# #############################################################################
#
# To Do:  1. Done.
#
#
# Done    1. Find way to generalize the criteria, to allow coverage of a wider
#            range of cases.

# #############################################################################

# def get_input_with_default(prompt, type, error_string, check_list =[], default = '', **kwds):
#     """
#     Return user input after validating it, offering the user a default value.
#     """
#
#     try:
#         prompt = 'CAS = [' + str(data_items['cas']) + '] '
#         value = VI.get_input(prompt, 'float+_or_blank',
#                        'ERROR - CAS must be a positive number.')
#         if value != '':
#             data_items['cas'] = value
#     except KeyError:
#         prompt = 'CAS = '
#         data_items['cas'] = VI.get_input(prompt, 'float+_or_blank',
#                        'ERROR - CAS must be a positive number.')
#
#     return data_items['cas']

"""
Validate user input against arbitrary criteria.  Used in the interactive interface in other modules.
"""


def get_input(prompt, type, error_string, check_list=[], default="", **kwds):
    """
    Return user input after validating it.
    """

    input_validated = False

    if type == "float":
        while not input_validated:
            input_data = input(prompt)
            try:
                input_data = float(input_data)
                input_validated = True
            except ValueError:
                print(error_string)

    if type == "float+":

        # data is a positive float

        input_data = input(prompt)
        try:
            input_data = float(input_data)
            if input_data < 0:
                print(error_string)
            else:
                input_validated = True
        except ValueError:
            print(error_string)
    elif type == "float_or_blank":

        # data is a float or blank

        while not input_validated:
            input_data = input(prompt)
            if input_data == "":
                input_data = default
                input_validated = True
            else:
                try:
                    input_data = float(input_data)
                    input_validated = True
                except ValueError:
                    print(error_string)
    elif type == "float_str_or_blank":

        # data is a float, a string in kwds['str_list'] or blank

        while not input_validated:
            input_data = input(prompt)
            if input_data == "":
                input_data = default
                input_validated = True
            elif input_data in kwds["str_list"]:
                input_validated = True
            else:
                try:
                    input_data = float(input_data)
                    input_validated = True
                except ValueError:
                    print(error_string)
    elif type == "float+_or_blank":

        # data is a float, >= 0, or blank

        while not input_validated:
            input_data = input(prompt)
            if input_data == "":
                input_data = default
                input_validated = True
                return input_data
            try:
                input_data = float(input_data)
                if input_data < 0:
                    print(error_string)
                else:
                    input_validated = True
            except ValueError:
                print(error_string)

    if type == "int":
        while 1:
            input_data = input(prompt)
            try:
                input_data = int(input_data)
                break
            except ValueError:
                print(error_string)

    if type == "int_or_str":

        # data is an int or a string in kwds['str_list']

        while 1:
            input_data = input(prompt)
            if input_data in kwds["str_list"]:
                break
            try:
                input_data = int(input_data)
                try:
                    if input_data < kwds["min"]:
                        print("You must enter an integer no less than", kwds["min"])
                except KeyError:
                    pass
                try:
                    if input_data > kwds["max"]:
                        print("You must enter an integer less than", kwds["max"])
                except KeyError:
                    break
            except ValueError:
                print("You must enter an integer")
    elif type == "list":

        while not input_validated:
            input_data = input(prompt)
            if input_data == "":
                input_data = default
                input_validated = True
                return default
            if input_data in check_list:
                input_validated = True
            else:
                print(error_string)

    return input_data


def get_input2(
    prompt, conditions_any=[], conditions_all=[], debug=False,
):
    """Return user input, after validating it against arbitrary criteria.

    The user input must satisfy all conditions in conditions_all, and one or
    more of the conditions in conditions_any.  Conditions_any is a list, with
    the first item the error string to display if none of the conditions are
    met.  Conditions_all is a list of tuples, with each tuple having one
    condition + the associated error string to display to the user if the
    condition is not met.

    The conditions will be inserted in 'if' statements, and must be written
    so that the 'if' statement will be true if the data is valid.  E.g.

    if a 'Q' is needed:
    'X == \"Q\"'

    Note 1: all data input by the user is seen as a string, even if numbers are
    entered.  If the test is looking for an integer, write it as:
    'int(X)'

    Note 2: All tests are run inside 'try' statements, so exceptions will not
    cause problems.  Any test that raises an exception is treated as a failed
    test.

    Note 3: One idiosyncrasy of 'if' statements is that any statement that
    returns a \"0\" is treated as a False.  Tests must be crafted so a pass
    does not return a '0'.  E.g., if the test is intended to check that the
    data is a float, and a zero would be valid data, do not write the test as:
    'float(X)', as 'if float(\"0\"):' will not pass.

    """

    input_validated = False

    while not input_validated:
        X = input(prompt)
        validated_any = False
        validated_all = True

        # does input meet one of conditions in conditions_any?

        if len(conditions_any) > 0:
            error_string = conditions_any[0]
            for condition in conditions_any[1:]:
                if debug:
                    print("Testing condition", condition)
                try:
                    if eval(condition):
                        validated_any = True
                        if debug:
                            print("Test of ", condition, "passed")
                    else:
                        if debug:
                            print("Test of ", condition, "failed")
                except:
                    if debug:
                        print("Exception during test")
                    pass

        # does input meet all conditions in conditions_all?

        if len(conditions_all) > 0:
            for condition in conditions_all:
                if debug:
                    print("Testing condition", condition[0])
                try:
                    if eval(condition[0]):
                        if debug:
                            print("Test of ", condition[0], "passed")
                        pass
                    else:
                        if debug:
                            print("Test of ", condition[0], "failed")
                        validated_all = False
                        # print(condition[1])
                except:
                    if debug:
                        print("Exception during test")
                    validated_all = False
                    # print(condition[1])

        if not validated_any:
            print(error_string, "\n")
        elif validated_all:
            input_validated = True
    return X



#!/usr/bin/python
# -*- coding: utf-8 -*-

# #############################################################################
# Copyright (c) 2008, Kevin Horton
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# *
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The name of Kevin Horton may not be used to endorse or promote products
#       derived from this software without specific prior written permission.
# *
# THIS SOFTWARE IS PROVIDED BY KEVIN HORTON ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL KEVIN HORTON BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# #############################################################################
#
# version 0.16, 06 May 2007
#
# Version History:
# vers     date     Notes
#  0.1   14 May 06  First release.
#
# 0.11   17 May 06  Cleaned up documentation.
#
# 0.12   24 May 06  Added temp2speed_of_sound function.
#
# 0.13   02 Jun 06  Added temp2isa and isa2temp functions.
#
# 0.14   22 Apr 07  Added density altitude vs temperature functions
#
# 0.15   29 Apr 07  Broke out sat_press and dry_press as public functions.
#
# 0.16   05 May 07  Reworked to use default units from default_units module.
# #############################################################################
#
# To Do: 1. Done.
#
#        2. Done.
#
#        3. Done.
#
#        4. Won't do.
#
#        5. Add temp2temp_ratio, press2press_ratio and density2density_ratio.
#
#        6. Done.
#
#        7. Add tests for all functions to test/test_std_atm.py
#           Tests to add:
#           dry_press
#           density_alt_table ? (probably won't make a test for this)
#
#        8. Review code to catch duplicated code blocks in different functions.
#           Move these blocks to internal functions.
#
#        9. Review API for all public functions for consistency of units, etc.
#
# Done:  1. consider replacing calculations by constants where possible.  See
#           http://jcoppens.com/globo/teoria/atm_est.en.php
#
#           Tested replacing calculations by constants in press2alt.  The
#           perf improvement was only about 15%.  Probably not worth the trouble.
#           Better to keep the pedigree of the equations visible.
#
#        2. Validate against published data.  Created unittests using data:
#           http://www.sworld.com.au/steven/space/atmosphere/
#
#        3. Added relative humidity to density altitude calc.  See:
#           http://wahiduddin.net/calc/density_altitude.htm
#
#        4. Change formulae to use pressure in pa, not in HG.  Won't do.
#           Instead, changed to use default units specified in default_units.py
#
#        6. Added functions:
#           isa2temp
#           temp2isa
#
#        7. Added tests for functions:
#           pressure_alt
#           sat_press
#           density_alt2temp
# #############################################################################

"""Calculate standard atmosphere parametres.

Calculates standard atmosphere parametres, using the 1976 International
Standard Atmosphere.  The default units for the input and output are defined
in default_units.py

All altitudes are geopotential altitudes (i.e. it is assumed that there is
no variation with altitude of the acceleration due to gravity).

Works up to 84.852 km (278,386 ft) altitude.

"""

from . import unit_conversion as U
import math as M
import locale as L

try:
    from .default_units import *
except ImportError:
    default_area_units = "ft**2"
    default_power_units = "hp"
    default_speed_units = "kt"
    default_temp_units = "C"
    default_weight_units = "lb"
    default_press_units = "in HG"
    default_density_units = "lb/ft**3"
    default_length_units = "ft"
    default_alt_units = default_length_units
    default_avgas_units = "lb"

try:
    L.setlocale(L.LC_ALL, "en_US")
except:
    pass

g = 9.80665  # Acceleration of gravity at 45.542 deg latitude, m/s**s
Rd = 287.05307  # Gas constant for dry air, J/kg K

# conditions starting at sea level, in a region with temperature gradient

T0 = 288.15  # Temperature at sea level, degrees K
L0 = -6.5  # Temperature lapse rate, at sea level deg K/km
P0 = 29.9213  # Pressure at sea level, in HG
Rho0 = 1.2250  # Density at sea level, kg/m**3

# conditions starting at 11 km, in an isothermal region

T11 = T0 + 11 * L0  # Temperature at 11,000 m, degrees K
PR11 = (T11 / T0) ** ((-1000 * g) / (Rd * L0))  # pressure ratio at 11,000 m
P11 = PR11 * P0
Rho11 = (Rho0 * PR11) * (T0 / T11)

# conditions starting at 20 km, in a region with temperature gradient

T20 = T11
PR20 = PR11 * M.exp(((-1000 * g) * (20 - 11)) / (Rd * T11))
L20 = 1  # temperature lapse rate, starting at 20,000 m, deg K/km
P20 = PR20 * P0
Rho20 = (Rho0 * PR20) * (T0 / T20)

# conditions starting at 32 km, in a region with temperature gradient

T32 = 228.65  # Temperature at 32 km, degrees K
PR32 = PR20 * (T32 / T20) ** ((-1000 * g) / (Rd * L20))

# PR32 = PR20 * M.exp((-1000 * g) * (32 - 20)/(R * T20))

L32 = 2.8  # temperature lapse rate, starting at 32,000 m, deg K/km
P32 = PR32 * P0
Rho32 = (Rho0 * PR32) * (T0 / T32)

# conditions starting at 47 km, in an isothermal region

T47 = 270.65
PR47 = PR32 * (T47 / T32) ** ((-1000 * g) / (Rd * L32))
P47 = PR47 * P0
Rho47 = (Rho0 * PR47) * (T0 / T47)

# conditions starting at 51 km, in a region with temperature gradient

T51 = 270.65  # Temperature at 51 km, degrees K
PR51 = PR47 * M.exp(((-1000 * g) * (51 - 47)) / (Rd * T47))
L51 = -2.8  # temperature lapse rate, starting at 51,000 m, deg K/km
P51 = PR51 * P0
Rho51 = (Rho0 * PR51) * (T0 / T51)

# conditions starting at 71 km, in a region with temperature gradient

T71 = 214.65  # Temperature at 71 km, degrees K
PR71 = PR51 * (T71 / T51) ** ((-1000 * g) / (Rd * L51))
L71 = -2.0  # temperature lapse rate, starting at 71,000 m, deg K/km
P71 = PR71 * P0
Rho71 = (Rho0 * PR71) * (T0 / T71)

# temp_units_list = ['C', 'F', 'K', 'R']

# #############################################################################
#
# Altitude to temperature
#
# #############################################################################


def alt2temp(H, alt_units=default_alt_units, temp_units=default_temp_units):
    """Return the standard temperature for the specified altitude.  Altitude
    units may be feet ('ft'), metres ('m'), statute miles, ('sm') or
    nautical miles ('nm').  Temperature units may be degrees C, F, K or R
    ('C', 'F', 'K' or 'R')

    If the units are not specified, the units in default_units.py are used.

    Examples:

    Calculate the standard temperature (in default temperature units) at
    5,000 (default altitude units):
    >>> alt2temp(5000)
    5.0939999999999941

    Calculate the standard temperature in deg F at sea level:
    >>> alt2temp(0, temp_units = 'F')
    59.0

    Calculate the standard temperature in deg K at 11,000 m:
    >>> alt2temp(11000, alt_units = 'm', temp_units = 'K')
    216.64999999999998

    Calculate the standard temperature at 11 statute miles in deg R:
    >>> alt2temp(11, alt_units = 'sm', temp_units = 'R')
    389.96999999999997

    The input value may be an expression:
    >>> alt2temp(11 * 5280, temp_units = 'R')
    389.96999999999997

    """

    # Validated to 84000 m
    # uses meters and degrees K for the internal calculations

    # function tested in tests/test_std_atm.py

    H = U.len_conv(H, from_units=alt_units, to_units="km")

    if H <= 11:
        temp = T0 + H * L0
    elif H <= 20:
        temp = T11
    elif H <= 32:
        temp = T20 + (H - 20) * L20
    elif H <= 47:
        temp = T32 + (H - 32) * L32
    elif H <= 51:
        temp = T47
    elif H <= 71:
        temp = T51 + (H - 51) * L51
    elif H <= 84.852:
        temp = T71 + (H - 71) * L71
    else:
        raise ValueError(
            "This function is only implemented for altitudes of 84.852 km and below."
        )

    return U.temp_conv(temp, to_units=temp_units, from_units="K")


def alt2temp_ratio(H, alt_units=default_alt_units):
    """
    Return the temperature ratio (temperature / standard temperature for
    sea level).  The altitude is specified in feet ('ft'), metres ('m'),
    statute miles, ('sm') or nautical miles ('nm').

    If the units are not specified, the units in default_units.py are used.

    Examples:

    Calculate the temperature ratio at 8,000 (default altitude units)
    >>> alt2temp_ratio(8000)
    0.94499531494013533

    Calculate the temperature ratio at 8,000 m.
    >>> alt2temp_ratio(8000, alt_units = 'm')
    0.81953843484296374
    """

    # function tested in tests/test_std_atm.py

    return alt2temp(H, alt_units, temp_units="K") / T0


# #############################################################################
#
# ISA deviation to temperature
#
# #############################################################################


def isa2temp(
    ISA_dev, altitude, temp_units=default_temp_units, alt_units=default_alt_units,
):
    """
    Return the temperature that is a specified amount warmer or cooler
    than the standard temperature for the altitude.

    The temperature may be in deg C, F, K or R.

    The altitude may be in feet ('ft'), metres ('m'), kilometres ('km'),
    statute miles, ('sm') or nautical miles ('nm').

    If the units are not specified, the units in default_units.py are used.

    Examples:

    Determine the temperature that is 10 deg (default temperature units) warmer
    than the standard temperature at 8,000 (default altitude units):
    >>> isa2temp(10, 8000)
    9.1503999999999905

    Determine the temperature that is 25 degrees K cooler than the standard
    temperature at 2000 m.
    >>> isa2temp(-25, 2000, temp_units = 'K', alt_units = 'm')
    250.14999999999998
    """

    # function tested in tests/test_std_atm.py

    temp = ISA_dev + alt2temp(altitude, alt_units, temp_units)

    return temp


# #############################################################################
#
# temperature to ISA deviation
#
# #############################################################################


def temp2isa(
    temp, altitude, temp_units=default_temp_units, alt_units=default_alt_units,
):
    """
    Return the amount that the specified temperature is warmer or cooler
    than the standard temperature for the altitude.

    The temperature may be in deg C, F, K or R.

    The altitude may be in feet ('ft'), metres ('m'), kilometres ('km'),
    statute miles, ('sm') or nautical miles ('nm').

    If the units are not specified, the units in default_units.py are used.

    Examples:

    Determine the ISA deviation for a temperature of 30 deg (default
    temperature units) at an altitude of 2000 (default altitude units):
    >>> temp2isa(30, 2000)
    18.962400000000002

    Determine the ISA deviation in degrees F for a temperature of 45 deg F
    at an altitude of 1000 m:
    >>> temp2isa(45, 1000, temp_units = 'F', alt_units = 'm')
    -2.2999999999999972
    """

    # function tested in tests/test_std_atm.py

    std_temp = alt2temp(altitude, alt_units, temp_units)
    ISA_dev = temp - std_temp

    return ISA_dev


# #############################################################################
#
# Altitude to pressure and pressure ratio
#
# #############################################################################


def _alt2press_ratio_gradient(
    H, Hb, Pb, Tb, L,
):

    # eqn from USAF TPS PEC binder, page PS1-31

    return (Pb / P0) * (1 + (L / Tb) * (H - Hb)) ** ((-1000 * g) / (Rd * L))


def _alt2press_ratio_isothermal(
    H, Hb, Pb, Tb,
):

    # eqn from USAF TPS PEC binder, page PS1-26

    return (Pb / P0) * M.exp((-1 * (H - Hb)) * ((1000 * g) / (Rd * Tb)))


def alt2press_ratio(H, alt_units=default_alt_units):
    """
    Return the pressure ratio (atmospheric pressure / standard pressure
    for sea level).  The altitude is specified in feet ('ft'), metres ('m'),
    statute miles, ('sm') or nautical miles ('nm').

    If the units are not specified, the units in default_units.py are used.

    Examples:

    Calculate the pressure ratio at 5000 (default altitude units):
    >>> alt2press_ratio(5000)
    0.8320481158727735

    Calculate the pressure ratio at 1000 m:
    >>> alt2press_ratio(1000, alt_units = 'm')
    0.88699304638887044

    The functions are only implemented at altitudes of 84.852 km and lower.
    >>> alt2press_ratio(90, alt_units = 'km')
    Traceback (most recent call last):
      File '<stdin>', line 1, in ?
      File './std_atm.py', line 189, in alt2press_ratio
    if H <= 20:
    ValueError: This function is only implemented for altitudes of 84.852 km and below.
    """

    # uses meters and degrees K for the internal calculations

    # function tested in tests/test_std_atm.py

    H = U.len_conv(H, from_units=alt_units, to_units="km")

    if H <= 11:
        return _alt2press_ratio_gradient(H, 0, P0, T0, L0)
    if H <= 20:
        return _alt2press_ratio_isothermal(H, 11, P11, T11)
    if H <= 32:
        return _alt2press_ratio_gradient(H, 20, P20, T20, L20)
    if H <= 47:
        return _alt2press_ratio_gradient(H, 32, P32, T32, L32)
    if H <= 51:
        return _alt2press_ratio_isothermal(H, 47, P47, T47)
    if H <= 71:
        return _alt2press_ratio_gradient(H, 51, P51, T51, L51)
    if H <= 84.852:
        return _alt2press_ratio_gradient(H, 71, P71, T71, L71)
    else:
        raise ValueError(
            "This function is only implemented for altitudes of 84.852 km and below."
        )


def alt2press(H, alt_units=default_alt_units, press_units=default_press_units):
    """
    Return the atmospheric pressure for a given altitude, with the
    altitude in feet ('ft'), metres ('m'), statute miles, ('sm') or nautical
    miles ('nm'), and the pressure in inches of HG ('in HG'), mm of HG
    ('mm HG'), psi, lb per sq. ft ('psf'), pa, hpa or mb.

    If the units are not specified, the units in default_units.py are used.

    Examples:

    Calculate the pressure in inches of mercury at 5,000 (default altitude
    units):
    >>> alt2press(5000)
    24.895961289464015

    Calculate the pressure in pounds per square foot at 10,000 (default
    altitude units):
    >>> alt2press(10000, press_units = 'psf')
    1455.3301392981359

    Calculate the pressure in pascal at 20 km:
    >>> alt2press(20, press_units = 'pa', alt_units = 'km')
    5474.8827144576408
    """

    # uses meters, inches of HG and degrees K for the internal calculations

    # function tested in tests/test_std_atm.py

    H = U.len_conv(H, from_units=alt_units, to_units="m")

    press = P0 * alt2press_ratio(H, alt_units="m")
    press = U.press_conv(press, from_units="in HG", to_units=press_units)

    return press


# #############################################################################
#
# Pressure altitude from barometric altitude and altimeter setting
#
# #############################################################################


def pressure_alt(H, alt_setting, alt_units=default_alt_units):
    """
    Return the pressure altitude, given the barometric altitude and the
    altimeter setting.

    Altimeter setting may have units of inches of HG, or hpa or mb.  If the
    altimeter setting value is less than 35, the units are assumed to be
    in HG, otherwise they are assumed to be hpa.  The altimeter setting must
    be in the range of 25 to 35 inches of mercury.

    The altitude may have units of feet ('ft'), metres ('m'), statute miles,
    ('sm') or nautical miles ('nm').

    If the units are not specified, the units in default_units.py are used.

    Examples:

    Calculate the pressure altitude for 1,000 (default altitude units)
    barometric altitude with altimeter setting of 30.92 in HG:
    >>> pressure_alt(1000, 30.92)
    88.612734282205338

    Calculate the pressure altitude for 1,000 (default altitude units)
    barometric altitude with altimeter setting of 1008 mb:
    >>> pressure_alt(1000, 1008)
    1143.6503495627171

    Calculate the pressure altitude in metres for 304.8 m barometric
    altitude with altimeter setting of 1008 mb:
    >>> pressure_alt(304.8, 1008, alt_units = 'm')
    348.58462654671621
    """

    H = U.len_conv(H, from_units=alt_units, to_units="ft")
    if alt_setting > 35:
        alt_setting = U.press_conv(alt_setting, from_units="hpa", to_units="in HG")
    if alt_setting < 25 or alt_setting > 35:
        raise ValueError("Altimeter setting out of range.")
    HP = H + 145442.2 * (1 - (alt_setting / P0) ** 0.190261)
    HP = U.len_conv(HP, from_units="ft", to_units=alt_units)
    return HP


def QNH(
    HP, H, alt_units=default_alt_units, alt_setting_units="in HG",
):
    """
    Return the altimeter setting, given the pressure altitude (HP) and the
    barometric altitude (H).
    """

    HP = U.len_conv(HP, from_units=alt_units, to_units="ft")
    H = U.len_conv(H, from_units=alt_units, to_units="ft")
    QNH = P0 * (1 - (HP - H) / 145442.2) ** 5.255594
    QNH = U.press_conv(QNH, from_units="in HG", to_units=alt_setting_units)

    return QNH


# #############################################################################
#
# Altitude to density and density ratio
#
# #############################################################################


def alt2density_ratio(H, alt_units=default_alt_units):
    """
    Return the density ratio (atmospheric density / standard density
    for sea level).  The altitude is specified in feet ('ft'), metres ('m'),
    statute miles, ('sm') or nautical miles ('nm').

    If the units are not specified, the units in default_units.py are used.

    Examples:

    Calculate the density ratio at 7,500 (default altitude units):
    >>> alt2density_ratio(7500)
    0.79825819881753035

    Calculate the density ratio at 2 km:
    >>> alt2density_ratio(2, alt_units = 'km')
    0.8216246960994622
    """

    # function tested in tests/test_std_atm.py

    return alt2press_ratio(H, alt_units) / alt2temp_ratio(H, alt_units)


def alt2density(H, alt_units=default_alt_units, density_units=default_density_units):
    """
    Return the density given the pressure altitude.  The altitude is
    specified in feet ('ft'), metres ('m'), statute miles, ('sm') or
    nautical miles ('nm').

    The desired density units are specified as 'lb/ft**3', 'slug/ft**3' or
    'kg/m**3'.

    If the units are not specified, the units in default_units.py are used.

    Examples:

    Calculate the density in lb / ft cubed at 7,500 (default altitude units):
    >>> alt2density(7500)
    0.061046199847730374

    Calculate the density in slugs / ft cubed at 5,000 (default altitude units):
    >>> alt2density(5000, density_units = 'slug/ft**3')
    0.0020480982157718704

    Calculate the density in kg / m cubed at 0 (default altitude units:
    >>> alt2density(0, density_units = 'kg/m**3')
    1.2250000000000001

    Calculate the density in kg / m cubed at 81,000 m:
    >>> alt2density(81000, density_units = 'kg/m**3', alt_units = 'm')
    1.3320480184052337e-05
    """

    # function tested in tests/test_std_atm.py

    # get density in kg/m**3

    density = Rho0 * alt2density_ratio(H, alt_units)
    return U.density_conv(density, from_units="kg/m**3", to_units=density_units)


# #############################################################################
#
# Density to altitude and density ratio to altitude
#
# #############################################################################


def _density2alt_gradient(
    Rho, Rhob, Hb, Tb, L,
):

    return Hb + (Tb / L) * ((Rho / Rhob) ** (-1 / ((1000 * g) / (Rd * L) + 1)) - 1)


def _density2alt_isothermal(
    Rho, Rhob, Hb, Tb,
):

    return Hb - ((Rd * Tb) * M.log(Rho / Rhob)) / (1000 * g)


def density2alt(Rho, density_units=default_density_units, alt_units=default_alt_units):
    """
    Return the altitude corresponding to the specified density, with
    density in 'lb/ft**3', 'slug/ft**3' or 'kg/m**3'.

    The altitude is specified in feet ('ft'), metres ('m'), statute miles,
    ('sm') or nautical miles ('nm').

    If the units are not specified, the units in default_units.py are used.

    Examples:

    Calculate the altitude in default altitude units where the density is
    0.056475 in default density units:
    >>> density2alt(.056475)
    9999.8040934937271

    Calculate the altitude in metres where the density is 0.018012 kg / m
    cubed:
    >>> density2alt(.018012, alt_units = 'm', density_units = 'kg/m**3')
    29999.978688508152
    """

    # function tested in tests/test_std_atm.py

    Rho = U.density_conv(Rho, from_units=density_units, to_units="kg/m**3")

    if Rho > Rho11:
        H = _density2alt_gradient(Rho, Rho0, 0, T0, L0)
    elif Rho > Rho20:
        H = _density2alt_isothermal(Rho, Rho11, 11, T11)
    elif Rho > Rho32:
        H = _density2alt_gradient(Rho, Rho20, 20, T20, L20)
    elif Rho > Rho47:
        H = _density2alt_gradient(Rho, Rho32, 32, T32, L32)
    elif Rho > Rho51:
        H = _density2alt_isothermal(Rho, Rho47, 47, T47)
    elif Rho > Rho71:
        H = _density2alt_gradient(Rho, Rho51, 51, T51, L51)
    else:
        H = _density2alt_gradient(Rho, Rho71, 71, T71, L71)

    if H > 84.852:
        raise ValueError(
            "This function is only implemented for altitudes of 84.852 km and below."
        )

    return U.len_conv(H, from_units="km", to_units=alt_units)


def density_ratio2alt(DR, alt_units=default_alt_units):
    """
    Return the altitude for the specified density ratio. The altitude is in
    feet ('ft'), metres ('m'), statute miles, ('sm') or nautical miles
    ('nm').

    If the units are not specified, the units in default_units.py are used.

    Examples:

    Calculate the altitude in default altitude units where the density ratio is
    1:
    >>> density_ratio2alt(1)
    0.0

    Calculate the altitude in feet where the density ratio is 0.5:
    >>> density_ratio2alt(.5)
    21859.50324995652

    Calculate the altitude in km where the density ratio is 0.1
    >>> density_ratio2alt(.1, alt_units = 'km')
    17.9048674520646
    """

    # function tested in tests/test_std_atm.py

    D = DR * Rho0
    return density2alt(D, alt_units=alt_units, density_units="kg/m**3")


# #############################################################################
#
# Density Altitude
#
# #############################################################################


def density_alt(
    H,
    T,
    alt_setting=P0,
    DP="FALSE",
    RH=0.0,
    alt_units=default_alt_units,
    temp_units=default_temp_units,
):
    """
    Return density altitude, given the pressure altitude and the
    temperature with altitudes in units of feet ('ft'), metres ('m'),
    statute miles, ('sm') or nautical miles ('nm'), and temperature in units
    of deg C, F, K or R ('C', 'F', 'K' or 'R').

    Mandatory parametres:
    H = altitude
    T = temperature

    Optional parametres:
    alt_setting = altimeter setting (defaults to 29.9213 if not provided
    DP = dew point
    RH = relative humidity
    alt_units = units for the altitude.  'ft', 'm', or 'km'.
    temp_units = units for the temperature and dew point.  'C', 'F', 'K'
                 or 'R'.

    The altimeter setting units are assumed to be inches of HG, unless the
    value is greater than 35.  In this case the units are assumed to be mb.

    If the dew point or relative humidity are not specified, the air is
    assumed to be completely dry.  If both the dew point and relative humidity
    are specified, the relative humidity value is ignored.

    If the units are not specified, the units in default_units.py are used.

    The method is from: http://wahiduddin.net/calc/density_altitude.htm

    Examples:

    Calculate the density altitude in default altitude units for a pressure
    altitude of 7000 default altitude units and a temperature of 15 deg
    (default temperature units).  The altimeter setting is not specified, so it
    defaults to standard pressure of 29.9213 in HG or 1013.25 mb:
    >>> density_alt(7000, 15)
    8595.3465863232504

    Calculate the density altitude in default altitude units for a pressure
    altitude of 7000 default altitude units and a temperature of 85 deg F.
    The altimeter setting is not specified, so it defaults to standard pressure
    of 29.9213 in HG or 1013.25 mb.  The dew point and relative humidity are
    not specified, so the air is assumed to be dry:
    >>> density_alt(7000, 85, temp_units = 'F')
    10159.10696106757

    Calculate the density altitude in default altitude units for a pressure
    altitude of 7000 default altitude units, an altimeter setting of 29.80 and
    a temperature of 85 deg F and a dew point of 55 deg F:
    >>> density_alt(7000, 85, 29.80, 55, temp_units = 'F')
    10522.776013011618

    Calculate the density altitude in metres for a pressure altitude of
    2000 m, an altimeter setting of 1010 mb,  a temperature of 15 deg (default
    temperature units) and a relative humidity of 50%:
    >>> density_alt(2000, 15, 1010, alt_units = 'm', RH = 0.5)
    2529.8230634449737

    The dew point may be specified in one of two ways: as the fourth
    argument on the command line, or via the keyword argument DP.
    >>> density_alt(2000, 15, 1010, alt_units = 'm', DP = 5)
    2530.7528237990618

    The relative humidity must be in the range of 0 to 1:
    >>> density_alt(2000, 15, 1010, alt_units = 'm', RH = 1.1)
    Traceback (most recent call last):
      File '<stdin>', line 1, in ?
      File 'std_atm.py', line 533, in density_alt
    raise ValueError, 'The relative humidity must be in the range of 0 to 1.'
    ValueError: The relative humidity must be in the range of 0 to 1.
    """

    Rv = 461.495  # gas constant for water vapour

    # saturated vapour pressure

    if DP == "FALSE" and RH == 0:
        Pv = 0
    else:
        Pv = sat_press(T, DP, RH, temp_units, press_units="pa")

    # dry air pressure

    Pd = dry_press(
        H, Pv, alt_setting=alt_setting, alt_units=alt_units, press_units="pa"
    )

    T = U.temp_conv(T, from_units=temp_units, to_units="K")
    D = Pd / (Rd * T) + Pv / (Rv * T)

    DR = D / Rho0

    return density_ratio2alt(DR, alt_units)


def _sat_press(T):
    """
    Return the saturation pressure in mb of the water vapour, given
    temperature in deg C.  Equation from:
    http://wahiduddin.net/calc/density_altitude.htm
    """

    eso = 6.1078
    c0 = 0.99999683
    c1 = -0.90826951e-2
    c2 = 0.78736169e-4
    c3 = -0.61117958e-6
    c4 = 0.43884187e-8
    c5 = -0.29883885e-10
    c6 = 0.21874425e-12
    c7 = -0.17892321e-14
    c8 = 0.11112018e-16
    c9 = -0.30994571e-19

    p = c0 + T * (
        c1
        + T
        * (
            c2
            + T * (c3 + T * (c4 + T * (c5 + T * (c6 + T * (c7 + T * (c8 + T * c9))))))
        )
    )
    sat_press = eso / p ** 8
    return sat_press


def sat_press(
    T="FALSE",
    DP="FALSE",
    RH=0.0,
    temp_units=default_temp_units,
    press_units=default_press_units,
):
    """
    Return the saturated vapour pressure of water.  Either the dew point, or
    the temperature and the relative humidity must be specified.  If both the
    dew point and relative humidity are specified, the relative humidity value
    is ignored.

    If the temperature and dew point are both specified, the dew point cannot
    be greater than the temperature:

    If the units are not specified, the units in default_units.py are used.

    >>> sat_press(T=10, DP=11)
    Traceback (most recent call last):
      File '<stdin>', line 1, in <module>
      File 'std_atm.py', line 795, in sat_press
        raise ValueError, 'The dew point cannot be greater than the temperature.'
    ValueError: The dew point cannot be greater than the temperature.

    Dew point is 11 deg (default temperature units).  Find the water vapour
    pressure in default pressure units:
    >>> sat_press(DP=11)
    0.38741015927568667

    Dew point is 65 deg F.  Find the water vapour pressure in default pressure units:
    >>> sat_press(DP=65, temp_units = 'F')
    0.62207710701956165

    Dew point is 212 deg F (the boiling point of water at sea level).
    Find the water vapour pressure in lb per sq. inch:
    >>> sat_press(DP=212, temp_units = 'F', press_units = 'psi')
    14.696764873564959

    Temperature is 30 deg C.  Find the water vapour pressure in default pressure units:
    for 50% relative humidity:
    >>> sat_press(T=30, RH = 0.5)
    0.62647666996057927
    """

    if DP != "FALSE":

        # use dew point method

        if T != "FALSE":
            if DP > T:
                raise ValueError(
                    "The dew point cannot be greater than the temperature."
                )

        DP = U.temp_conv(DP, from_units=temp_units, to_units="C")

        # calculate vapour pressure

        Pv = _sat_press(DP) * 100
    else:

        if RH == "FALSE":
            raise ValueError(
                "Either DP (dew point) or RH (relative humidity) must be specified."
            )

        # relative humidity is specified
        # confirm relative humidity is in range

        if RH < 0 or RH > 1:
            raise ValueError("The relative humidity must be in the range of 0 to 1.")

        if T == "FALSE":
            raise ValueError(
                "If the relative humidity is specified, the temperature must also be specified."
            )

        T = U.temp_conv(T, from_units=temp_units, to_units="C")

        Pv = _sat_press(T) * 100
        Pv *= RH

    Pv = U.press_conv(Pv, from_units="pa", to_units=press_units)

    return Pv


def dry_press(
    H, Pv, alt_setting=P0, alt_units=default_alt_units, press_units=default_press_units,
):
    """
    Returns dry air pressure, i.e. the total air pressure, less the water
    vapour pressure.
    """

    HP = pressure_alt(H, alt_setting, alt_units=alt_units)
    P = alt2press(HP, press_units=press_units, alt_units=alt_units)
    Pd = P - Pv

    return Pd


def density_alt2temp(
    density_alt_seek,
    press_alt,
    alt_units=default_alt_units,
    temp_units=default_temp_units,
):
    """
    Return temperature to achieve a desired density altitude.

    If the units are not specified, the units in default_units.py are used.
    """

    low = -100  # initial lower guess
    high = 100  # initial upper guess

    # confirm initial low and high are OK:

    da_low = density_alt(press_alt, low, alt_units=alt_units)
    if da_low > density_alt_seek:
        raise ValueError("Initial low guess too high.")

    da_high = density_alt(press_alt, high, alt_units=alt_units)
    if da_high < density_alt_seek:
        raise ValueError("Initial high guess too low.")

    guess = (low + high) / 2.0
    da_guess = density_alt(press_alt, guess, alt_units=alt_units)

    # keep iterating until da is within 1 ft of desired value

    while M.fabs(da_guess - density_alt_seek) > 1:
        if da_guess > density_alt_seek:
            high = guess
        else:
            low = guess

        guess = (low + high) / 2.0
        da_guess = density_alt(press_alt, guess, alt_units=alt_units)

    guess = U.temp_conv(guess, from_units="C", to_units=temp_units)

    return guess


def density_alt_table(
    density_alt_seek,
    alt_range=2000,
    alt_inc=100,
    alt_units=default_alt_units,
    temp_units=default_temp_units,
    multi_units=False,
    file="",
    format="text",
):
    """
    Return a text or html table of required temperature vs pressure altitude.

    If the units are not specified, the units in default_units.py are used.
    """

    line_buffer = []
    if format == "text":
        line_buffer.append("Pressure altitudes and temperatures for a density ")
        line_buffer.append("altitude of " + str(density_alt_seek) + " " + alt_units)
        line_buffer.append("(assuming dry air)\n")
        if multi_units:
            line_buffer.append(" Pressure    Temp      Temp")
            line_buffer.append(" Altitude")
            line_buffer.append("   (" + alt_units + ")     (deg C)   (deg F)")
        else:
            line_buffer.append(" Pressure    Temp")
            line_buffer.append(" Altitude")
            line_buffer.append("   (" + alt_units + ")     (deg " + temp_units + ")")
    elif format == "html":
        print("creating html")
    else:
        raise ValueError('Invalid format.  Must be either "text" or "html"')

    if multi_units:
        for alt in range(
            max(density_alt_seek - alt_range / 2.0, 0),
            density_alt_seek + alt_range / 2.0 + alt_inc,
            alt_inc,
        ):
            temp_c = density_alt2temp(density_alt_seek, alt, alt_units=alt_units)
            temp_f = U.temp_conv(temp_c, from_units="C", to_units="F")
            alt_str = L.format("%.*f", (0, alt), grouping=True)
            temp_c_str = "%.1f" % temp_c
            temp_f_str = "%.1f" % temp_f
            line_buffer.append(
                alt_str.rjust(6) + temp_c_str.rjust(11) + temp_f_str.rjust(10)
            )
    else:
        for alt in range(
            max(density_alt_seek - alt_range / 2.0, 0),
            density_alt_seek + alt_range / 2.0 + alt_inc,
            alt_inc,
        ):
            alt_str = L.format("%.*f", (0, alt), grouping=True)
            temp_str = "%.1f" % density_alt2temp(
                density_alt_seek, alt, temp_units=temp_units, alt_units=alt_units
            )
            line_buffer.append(alt_str.rjust(6) + temp_str.rjust(11))

    if file != "":
        OUT = open(file, "w")
        for line in line_buffer:
            OUT.write(line + "\n")

        # print("file selected")
    else:
        return "\n".join(line_buffer)


# #############################################################################
#
# Pressure to altitude and pressure ratio to altitude
#
# #############################################################################


def _press2alt_gradient(
    P, Pb, Hb, Tb, L,
):

    return Hb + (Tb / L) * ((P / Pb) ** (((-1 * Rd) * L) / (1000 * g)) - 1)


def _press2alt_isothermal(
    P, Pb, Hb, Tb,
):

    return Hb - ((Rd * Tb) * M.log(P / Pb)) / (1000 * g)


def press2alt(P, press_units=default_press_units, alt_units=default_alt_units):
    """
    Return the altitude corresponding to the specified pressure, with
    pressure in inches of HG, mm of HG, psi, psf (lb per sq. ft), pa, hpa or
    mb.

    The altitude is in units of feet ('ft'), metres ('m'), statute miles,
    ('sm') or nautical miles ('nm')

    If the units are not specified, the units in default_units.py are used.

    Examples:

    Calculate the pressure altitude in feet for a pressure of 31.0185 inches
    of HG:
    >>> press2alt(31.0185)
    -999.98992888235091

    Calculate the pressure altitude in feet for a pressure of
    1455.33 lb sq. ft:
    >>> press2alt(1455.33, press_units = 'psf')
    10000.002466564831

    Calculate the pressure altitude in metres for a pressure of
    90.3415 mm HG:
    >>> press2alt(90.3415, press_units = 'mm HG', alt_units = 'm')
    15000.025465320754

    Calculate the pressure altitude in metres for a pressure of
    1171.86 pascal:
    >>> press2alt(1171.86, press_units = 'pa', alt_units = 'm')
    30000.029510365184
    """

    # function tested in tests/test_std_atm.py

    P = U.press_conv(P, from_units=press_units, to_units="in HG")

    if P > P11:
        H = _press2alt_gradient(P, P0, 0, T0, L0)
    elif P > P20:
        H = _press2alt_isothermal(P, P11, 11, T11)
    elif P > P32:
        H = _press2alt_gradient(P, P20, 20, T20, L20)
    elif P > P47:
        H = _press2alt_gradient(P, P32, 32, T32, L32)
    elif P > P51:
        H = _press2alt_isothermal(P, P47, 47, T47)
    elif P > P71:
        H = _press2alt_gradient(P, P51, 51, T51, L51)
    else:
        H = _press2alt_gradient(P, P71, 71, T71, L71)

    if H > 84.852:
        raise ValueError(
            "This function is only implemented for altitudes of 84.852 km and below."
        )

    return U.len_conv(H, from_units="km", to_units=alt_units)


def press_ratio2alt(PR, alt_units=default_alt_units):
    """
    Return the pressure ratio for the specified altitude.  The altitude is
    specified in feet ('ft'), metres ('m'), statute miles, ('sm') or
    nautical miles ('nm').

    If the units are not specified, the units in default_units.py are used.

    Examples:

    Calculate the altitude in feet where the pressure ratio is 0.5:
    >>> press_ratio2alt(.5)
    17969.990746028907

    Calculate the altitude in metres where the pressure ratio is 0.1:
    >>> press_ratio2alt(.1, alt_units = 'm')
    16096.249927559489
    """

    # function tested in tests/test_std_atm.py

    P = PR * P0
    return press2alt(P, alt_units=alt_units)


# #############################################################################
#
# Temperature to speed of sound
#
# #############################################################################


def temp2speed_of_sound(
    temp, temp_units=default_temp_units, speed_units=default_speed_units
):
    """
    Return the speed of sound, given the air temperature.

    The temperature units may be deg C, F, K or R ('C', 'F', 'K' or 'R').

    The speed units may be 'kt', 'mph', 'km/h', 'm/s' and 'ft/s'.

    If the units are not specified, the units in default_units.py are used.

    Examples:

    Determine speed of sound in knots at 15 deg (default temperature units):
    >>> temp2speed_of_sound(15)
    661.47882487301808

    Determine speed of sound in mph at 120 deg F:
    >>> temp2speed_of_sound(120, speed_units = 'mph', temp_units = 'F')
    804.73500154991291
    """

    # function tested in tests/test_std_atm.py

    temp = U.temp_conv(temp, from_units=temp_units, to_units="K")

    speed_of_sound = M.sqrt((1.4 * Rd) * temp)
    speed_of_sound = U.speed_conv(
        speed_of_sound, from_units="m/s", to_units=speed_units
    )

    return speed_of_sound


if __name__ == "__main__":

    # run doctest to check the validity of the examples in the doc strings.

    import doctest
    import sys

    doctest.testmod(sys.modules[__name__])



#!/usr/bin/env python
# encoding: utf-8

# #############################################################################
# Copyright (c) 2008, Kevin Horton
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# *
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The name of Kevin Horton may not be used to endorse or promote products
#       derived from this software without specific prior written permission.
# *
# THIS SOFTWARE IS PROVIDED BY KEVIN HORTON ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL KEVIN HORTON BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# #############################################################################
#
# version 0.10, 25 Apr 2008
#
# Version History:
# vers     date     Notes
# 0.10   25 Apr 08  First public release.
# #############################################################################

"""
Various functions related to static source error correction.
"""

from . import airspeed as A
import math as M
from . import std_atm as SA
from . import unit_conversion as U

##############################################################################
#
# TAS from GPS data.
#
##############################################################################
def gps2tas(GS, TK, verbose=0):
    """
    Returns true airspeed, given GPS groundspeed and track on at least
    three legs (four legs preferred).  Uses the method developed by Doug
    Gray - http://www.kilohotel.com/rv8/rvlinks/doug_gray/TAS_FNL4.pdf

    GS and TK are lists of ground speed and track data.

    Three legs:
        If verbose = 0, then only TAS is returned.
        If verbose = 1, then TAS, wind speed and direction are returned.
        If verbose = 2, then TAS, wind speed and direction and the heading
        for each leg are returned.  The wind speed and direction is
        returned as a tuple, and the headings are returned as a tuple


    Four legs:
        Data from only three legs is sufficient to calculate TAS.  If data
        four legs is entered, four different calculations are conducted,
        using a different mix of three data points for each calculation.
        If the data quality is high, the TAS and wind for all four
        calculations will be similar.  The standard deviation on the TAS is
        calculated - good quality data will have a standard deviation of
        less than 1 kt.

        If verbose = 0, then only TAS is returned.
        If verbose = 1, then TAS and its standard deviation are returned.
        If verbose = 2, then TAS, its standard deviation and the four wind
        speeds and directions are returned (the winds are returned as a list
        of four tuples)

    Validated against sample data in four leg tab of NTPS GPS PEC spreadsheet:
    http://www.ntps.edu/Files/GPS%20PEC.XLS

    Examples:

        Data for all examples:
            >>> gs = [178, 185, 188, 184]
            >>> tk = [178, 82, 355, 265]

            Determine the TAS, given the above data from four runs:
            >>> gps2tas(gs, tk)
            183.72669557114619

            Determine the TAS and standard deviation from the four calculations:
            >>> gps2tas(gs, tk, verbose = 1)
            (183.72669557114619, 0.82709634705928436)

            Determine the TAS, standard deviation, and wind speed and direction
            for each calculation:
            >>> gps2tas(gs, tk, verbose = 2)
            (183.72669557114619, 0.82709634705928436, ((5.2608369270843056, 194.51673740323213), (3.5823966532035927, 181.52174627838372), (5.1495218164839995, 162.69803415599802), (6.4436728241320145, 177.94783081049718)))

    """
    # confirm GS and TK are valid lengths:
    if 2 < len(GS) < 5:
        pass
    else:
        raise ValueError("GS must be a list of three or four items")

    if 2 < len(TK) < 5:
        pass
    else:
        raise ValueError("TK must be a list of three or four items")

    if len(GS) != len(TK):
        raise ValueError(
            "The ground speed and track arrays must have the same number of elements."
        )

    if len(GS) == 3:
        result = gps2tas3(GS, TK, verbose)
        return result
    else:
        gs_data_sets, tk_data_sets, results = [], [], []

        gs_data_sets.append([GS[0], GS[1], GS[2]])
        gs_data_sets.append([GS[1], GS[2], GS[3]])
        gs_data_sets.append([GS[2], GS[3], GS[0]])
        gs_data_sets.append([GS[3], GS[0], GS[1]])

        tk_data_sets.append([TK[0], TK[1], TK[2]])
        tk_data_sets.append([TK[1], TK[2], TK[3]])
        tk_data_sets.append([TK[2], TK[3], TK[0]])
        tk_data_sets.append([TK[3], TK[0], TK[1]])

        for (gs, tk) in zip(gs_data_sets, tk_data_sets):
            results.append(gps2tas3(gs, tk, 2))

        ave_TAS = 0
        ave_wind_x = 0
        ave_wind_y = 0
        sum2_TAS = 0

        for item in results:
            ave_TAS += item[0]
            sum2_TAS += item[0] ** 2
            ave_wind_x += item[1][0] * M.sin(M.pi * item[1][1] / 180.0)
            ave_wind_y += item[1][0] * M.cos(M.pi * item[1][1] / 180.0)

        ave_TAS /= 4.0
        std_dev_TAS = M.sqrt((sum2_TAS - 4 * ave_TAS ** 2) / 3)
        ave_wind_x /= 4
        ave_wind_y /= 4.0
        ave_wind_speed = M.sqrt(ave_wind_x ** 2 + ave_wind_y ** 2)
        ave_wind_dir = (720.0 - (180.0 / M.pi * M.atan2(ave_wind_x, ave_wind_y))) % 360
        # return results

        if verbose == 0:
            return ave_TAS
        elif verbose == 1:
            return ave_TAS, std_dev_TAS
        elif verbose == 2:
            return (
                ave_TAS,
                std_dev_TAS,
                (
                    (results[0][1][0], results[0][1][1]),
                    (results[1][1][0], results[1][1][1]),
                    (results[2][1][0], results[2][1][1]),
                    (results[3][1][0], results[3][1][1]),
                ),
            )
        else:
            raise ValueError("The value of verbose must be equal to 0, 1 or 2")


def gps2tas3(GS, TK, verbose=0):
    """
    Returns true airspeed, given GPS groundspeed and track on three legs.
    Uses the method developed by Doug Gray:
    http://www.kilohotel.com/rv8/rvlinks/doug_gray/TAS_FNL4.pdf

    GS and TK are arrays of ground speed and track data.

    If verbose = 0, then only TAS is returned.
    If verbose = 1, then TAS, wind speed and wind direction are returned.
    If verbose = 2, then TAS, wind speed and direction and the heading for
    each leg are returned.  The wind speed and direction is returned as a
    tuple, and the headings are returned as a tuple

    Validated against sample in Doug Gray's paper.

    Examples:

    """
    x, y, b, m, hdg = [], [], [], [], []

    for (gs, tk) in zip(GS, TK):
        x.append(gs * M.sin(M.pi * (360.0 - tk) / 180.0))
        y.append(gs * M.cos(M.pi * (360.0 - tk) / 180.0))

    m.append(-1 * (x[1] - x[0]) / (y[1] - y[0]))
    m.append(-1 * (x[2] - x[0]) / (y[2] - y[0]))

    b.append((y[0] + y[1]) / 2 - m[0] * (x[0] + x[1]) / 2)
    b.append((y[0] + y[2]) / 2 - m[1] * (x[0] + x[2]) / 2)

    wind_x = (b[0] - b[1]) / (m[1] - m[0])
    wind_y = m[0] * wind_x + b[0]

    wind_speed = M.sqrt(wind_x ** 2 + wind_y ** 2)
    wind_dir = (540.0 - (180.0 / M.pi * M.atan2(wind_x, wind_y))) % 360.0

    TAS = M.sqrt((x[0] - wind_x) ** 2 + (y[0] - wind_y) ** 2)

    if verbose >= 2:
        hdg.append(
            (540.0 - (180.0 / M.pi * M.atan2(wind_x - x[0], wind_y - y[0]))) % 360.0
        )
        hdg.append(
            (540.0 - (180.0 / M.pi * M.atan2(wind_x - x[1], wind_y - y[1]))) % 360.0
        )
        hdg.append(
            (540.0 - (180.0 / M.pi * M.atan2(wind_x - x[2], wind_y - y[2]))) % 360.0
        )

        return TAS, (wind_speed, wind_dir), (hdg[0], hdg[1], hdg[2])

    elif verbose == 1:
        return TAS, (wind_speed, wind_dir)
    elif verbose == 0:
        return TAS
    else:
        raise ValueError("The value of verbose must be equal to 0, 1 or 2")


def main():
    pass


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# #############################################################################
# Copyright (c) 2008, Kevin Horton
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# *
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The name of Kevin Horton may not be used to endorse or promote products
#       derived from this software without specific prior written permission.
# *
# THIS SOFTWARE IS PROVIDED BY KEVIN HORTON ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL KEVIN HORTON BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# #############################################################################
#
# version 0.27, 16 May 2007
#
# Version History:
# vers     date      Notes
# 0.21   10 June 06  Started to add interactive mode
# 0.22   10 July 06  Error correction in some default units
# 0.23   07 Jan 07   Added several more interactive functions
# 0.24   05 May 07   Reworked to use default units from default_units module.
# 0.25   06 May 07   Added several more interactive functions
# 0.26   12 May 07   Reworked the interactive mode to move common idioms to
#                    functions
# 0.27   16 May 07   Reworked to respect default units, to stay open after
#                    running, and to save results between functions.
# #############################################################################
#
# To Do:  1. Add functions:
#
#         2. Extend the following functions to work at M > 1 and CAS > 661.48:
#             dp_over_p2mach
#             dp2eas
#             dp2tas
#             eas2dp
#             tas2dp
#
#         3. Test the following functions:
#             cas2eas
#             eas2cas
#             cas_mach2alt
#             cas_alt2mach
#             mach_alt2cas
#             mach2tas # errors?
#             tas2mach # errors?
#
#         4. Add examples to all functions.
#
#         5. Add tests for all functions.
#
#         6. Check superconic versions of cas to mach and mach to cas equations
#            from TPS notes
#
#         7. Done.
#
#         8. Done.
#
#
# Done    7. Rework the interactive functions to respect the default units.
#
#         8. Rework interactive functions to share data between runs.

# #############################################################################
# % coverage.py -r -m ../airspeed.py
# Name          Stmts   Exec  Cover   Missing
# -------------------------------------------
# ../airspeed     422     52    12%   98-109, 112, 115-117, 122-127, 156-198, 207, 217, 221, 223, 226, 231-278, 310, 315-316, 346-365, 370-404, 435-436, 440-444, 479-521, 530-600, 603-634, 660-664, 674-704, 707-722, 750-889, 891-892, 903-904, 931, 945-989, 998-1050, 1053-1209, 1231-1299, 1360-1362, 1368-1389, 1395, 1398, 1400-1401, 1411-1503, 1513-1530, 1562-1604
"""
Perform various air speed conversions.

Convert between Indicated Air Speed (IAS), Calibrated Air Speed (CAS),
Equivalent Air Speed (EAS), True Air Speed (TAS) and mach number.

Convert between pitot static system pressures and air speed.

Provide interactive airspeed conversions when script is run directly, e.g.
'python airspeed.py'.

"""

import math as M
from . import std_atm as SA

try:
    from .default_units import *
except ImportError:
    default_area_units = "ft**2"
    default_power_units = "hp"
    default_speed_units = "kt"
    default_temp_units = "C"
    default_weight_units = "lb"
    default_press_units = "in HG"
    default_density_units = "lb/ft**3"
    default_length_units = "ft"
    default_alt_units = default_length_units
    default_avgas_units = "lb"

from . import unit_conversion as U
from . import val_input as VI

Rho0 = SA.Rho0  # Density at sea level, kg/m**3
P0 = 101325.0  # Pressure at sea level, pa

# speed of sound from http://www.edwards.af.mil/sharing/tech_pubs/Handbook-10%20March02.pdf

A0 = 340.2941  # speed of sound at sea level, std day, m/s

# F calculated by manipulating NASA RP 1046 pg 17
# F is used in some of the supersonic solution equations.

F = (1.25 ** 2.5 * (2.4 ** 2.0) ** 2.5) * 1.2

# #############################################################################
#
# delta pressure to speed
#
#    delta pressure to CAS
#
#    delta pressure and altitude to EAS
#
#    delta pressure, altitude and temperature to TAS
#
# #############################################################################


def _dp2speed(
    dp, Pref, Rhoref, press_units=default_press_units, speed_units=default_speed_units,
):

    dp = U.press_conv(dp, from_units=press_units, to_units="pa")
    speed = M.sqrt(
        ((7.0 * Pref) * (1.0 / Rhoref)) * ((dp / Pref + 1.0) ** (2.0 / 7.0) - 1.0)
    )

    # check to confirm the speed is less than 661.48 kt

    speed_kt = U.speed_conv(speed, from_units="m/s", to_units="kt")
    if speed_kt > 661.48:
        raise ValueError(
            "The function _dp2speed only works if the speed is less than or equal to 661.48 kt"
        )
    speed = U.speed_conv(speed, from_units="m/s", to_units=speed_units)

    return speed


def dp2cas(dp, press_units=default_press_units, speed_units=default_speed_units):
    """
    Return the CAS for a given differential pressure (the difference
    between the pitot and static pressures).

    The pressure units may be in inches of HG, mm of HG, psi, lb/ft^2,
    hpa and mb.  The units are specified as: 'in HG', 'mm HG', 'psi',
    'lb/in**2', 'psf', 'lb/ft**2 'hpa', 'mb' or 'pa'.

    The speed units may be 'kt', 'mph', 'km/h', 'm/s' and 'ft/s'.

    If the units are not specified, the units in default_units.py are used.

    Examples:

    Determine the CAS in kt that is equivalent to a differential pressure
    of 15 in HG:
    >>> dp2cas(15)
    518.96637566127652

    Determine the CAS in mph that is equivalent to a differential pressure
    of 0.2 psi:
    >>> dp2cas(.2, press_units = 'psi', speed_units = 'mph')
    105.88275188221526
    """

    try:

        # subsonic case

        cas = _dp2speed(dp, P0, Rho0, press_units, speed_units)
        return cas
    except ValueError:

        # supersonic case - need to iterate a solution.  Set upper and lower
        # guesses, and iterate until we zero in on a cas that produces the
        # desired differential pressure.

        dp_seek = U.press_conv(dp, from_units=press_units, to_units="pa")

        low = 340  # initial lower guess, m/s

        # This function works up to approximately 6,600 kt CAS.  The upper
        # limit can be extended by increasing the value of the initial upper
        # guess ("high").

        high = 3400  # initial upper guess, m/s

        # confirm initial low and high are OK:

        dp_low = _super_cas2dp(low)
        if dp_low > dp_seek:
            raise ValueError("Initial lower cas guess is too high.")

        dp_high = _super_cas2dp(high)
        if dp_high < dp_seek:
            raise ValueError("Initial upper cas guess is too low.")

        guess = (low + high) / 2.0
        dp_guess = _super_cas2dp(guess)

        # keep iterating until dp is within 0.001% of desired value

        while M.fabs(dp_guess - dp_seek) / dp_seek > 1e-5:
            if dp_guess > dp_seek:
                high = guess
            else:
                low = guess

            guess = (low + high) / 2.0
            dp_guess = _super_cas2dp(guess)

    # the supersonic solution is in m/s, so we need to convert to the
    # desired units.

    cas = U.speed_conv(guess, from_units="m/s", to_units=speed_units)

    # Supersonic case.  The subsonic case is returned way up at the top,
    # inside the try statement.

    return cas


def dp2eas(
    dp,
    altitude,
    press_units=default_press_units,
    speed_units=default_speed_units,
    alt_units=default_alt_units,
):
    """
    Return the EAS for a given differential pressure (the difference
    between the pitot and static pressures) and altitude.

    The pressure units may be in inches of HG, mm of HG, psi, lb/ft^2,
    hpa and mb.  The units are specified as: 'in HG', 'mm HG', 'psi',
    'lb/in**2', 'psf', 'lb/ft**2 'hpa', 'mb' or 'pa'.

    The speed units may be 'kt', 'mph', 'km/h', 'm/s' and 'ft/s'.

    The altitude may be in feet ('ft'), metres ('m'), kilometres ('km'),
    statute miles, ('sm') or nautical miles ('nm').

    If the units are not specified, the units in default_units.py are used.

    This first version only works for EAS < 661.48 kt.
    """

    P = SA.alt2press(altitude, alt_units, press_units="pa")

    eas = _dp2speed(dp, P, Rho0, press_units, speed_units)
    return eas


def dp2tas(
    dp,
    altitude,
    temp,
    press_units=default_press_units,
    speed_units=default_speed_units,
    alt_units=default_alt_units,
    temp_units=default_temp_units,
):
    """
    Return the TAS for a given differential pressure (the difference
    between the pitot and static pressures) and altitude.

    The pressure units may be in inches of HG, mm of HG, psi, lb/ft^2,
    hpa and mb.  The units are specified as: 'in HG', 'mm HG', 'psi',
    'lb/in**2', 'psf', 'lb/ft**2 'hpa', 'mb' or 'pa'.

    The speed units may be 'kt', 'mph', 'km/h', 'm/s' and 'ft/s'.

    The altitude may be in feet ('ft'), metres ('m'), kilometres ('km'),
    statute miles, ('sm') or nautical miles ('nm').

    The temperature may be in deg C, F, K or R.

    If the units are not specified, the units in default_units.py are used.

    This first version only works for TAS < 661.48 kt.
    """

    P = SA.alt2press(altitude, alt_units, press_units="pa")

    press_ratio = SA.alt2press_ratio(altitude, alt_units)
    temp_ratio = U.temp_conv(temp, from_units=temp_units, to_units="K") / 288.15
    density_ratio = press_ratio / temp_ratio
    Rho = Rho0 * density_ratio

    tas = _dp2speed(dp, P, Rho, press_units, speed_units)
    return tas


# #############################################################################
#
# speed to delta pressure
#
#    CAS to delta pressure
#
#    EAS and altitude to delta pressure
#
#    TAS, altitude and temperature to delta pressure
#
# #############################################################################


def _speed2dp(
    speed,
    Pref,
    Rhoref,
    press_units=default_press_units,
    speed_units=default_speed_units,
):
    """ Return a delta pressure (the difference between the pitot and
    static pressures) for a given speed.  Subsonic equation.
    """

    speed = U.speed_conv(speed, from_units=speed_units, to_units="m/s")
    dp = Pref * (((Rhoref * speed ** 2.0) / (7.0 * Pref) + 1.0) ** 3.5 - 1.0)
    dp = U.press_conv(dp, from_units="pa", to_units=press_units)

    return dp


def _super_cas2dp(mcas: float) -> float:
    """Return the  differential pressure (difference between pitot and static
    pressures) for a given CAS.
    This function  only  works for the following conditions:
        - CAS > 661.48 kt
        - speed in m/s
        - pressure in pa.

    Args:
        mcas (float): [description]
        #TODO: describe mcas

    Returns:
        float : differential pressure (difference between pitot and static
    pressures) for a given CAS.
    """

    dp_over_P0 = (F * (mcas / A0) ** 7.0) / (
        7.0 * (mcas / A0) ** 2.0 - 1.0
    ) ** 2.5 - 1.0
    dp = dp_over_P0 * P0

    return dp


def cas2dp(
    cas, speed_units=default_speed_units, press_units=default_press_units
) -> float:
    """Converts calibrated airspeed to differential pressure (difference between pitot and static
    pressures) for a given CAS

    Args:
        cas (float): calibrated airspeed
        speed_units (string, optional): kt', 'mph', 'km/h', 'm/s' or 'ft/s'. Defaults to default_speed_units.
        press_units (str, optional): inches of HG, mm of HG, psi, lb/ft^2,
                                hpa and mb.
                                The units are specified as: 'in HG',
                                'mm HG', 'psi','lb/in**2', 'psf', 'lb/ft**2 'hpa',
                                'mb' or 'pa'.
                                 Defaults to default_press_units.
    If the units are not specified, the units in default_units.py are used.

    Returns:
        float: differential pressure (difference between pitot and static
    pressures) for a given CAS
    """

    # check to confirm the speed is less than 661.48 kt
    #   kcas = U.speed_conv(cas, from_units = speed_units, to_units = 'kt')

    mcas = U.speed_conv(cas, from_units=speed_units, to_units="m/s")

    #   if kcas > 661.48:

    if mcas > A0:

        # supersonic case

        dp = _super_cas2dp(mcas)
        dp = U.press_conv(dp, from_units="pa", to_units=press_units)
    else:

        # subsonic case

        dp = _speed2dp(cas, P0, Rho0, press_units=press_units, speed_units=speed_units)
    return dp


def eas2dp(
    eas,
    altitude,
    speed_units=default_speed_units,
    alt_units=default_alt_units,
    press_units=default_press_units,
):
    """
    Return the differential pressure (difference between pitot and static
    pressures) for a given EAS.

    The speed units may be 'kt', 'mph', 'km/h', 'm/s' and 'ft/s'.

    The pressure units may be in inches of HG, mm of HG, psi, lb/ft^2,
    hpa and mb.  The units are specified as: 'in HG', 'mm HG', 'psi',
    'lb/in**2', 'psf', 'lb/ft**2 'hpa', 'mb' or 'pa'.

    The altitude may be in feet ('ft'), metres ('m'), kilometres ('km'),
    statute miles, ('sm') or nautical miles ('nm').

    If the units are not specified, the units in default_units.py are used.

    This first version only works for CAS < 661.48 kt.

    """

    # check to confirm the speed is less than 661.48 kt

    keas = U.speed_conv(eas, from_units=speed_units, to_units="kt")
    if keas > 661.48:
        raise ValueError(
            "The function eas2dp only works if the eas is less than or equal to 661.48 kt"
        )

    P = SA.alt2press(altitude, alt_units=alt_units, press_units="pa")
    dp = _speed2dp(eas, P, Rho0, press_units=press_units, speed_units=speed_units)

    return dp


def tas2dp(
    tas,
    altitude,
    temp,
    speed_units=default_speed_units,
    alt_units=default_alt_units,
    temp_units=default_temp_units,
    press_units=default_press_units,
):
    """
    Return the differential pressure (difference between pitot and static
    pressures) for a given TAS.

    The speed units may be 'kt', 'mph', 'km/h', 'm/s' and 'ft/s'.

    The pressure units may be in inches of HG, mm of HG, psi, lb/ft^2,
    hpa and mb.  The units are specified as: 'in HG', 'mm HG', 'psi',
    'lb/in**2', 'psf', 'lb/ft**2 'hpa', 'mb' or 'pa'.

    The altitude may be in feet ('ft'), metres ('m'), kilometres ('km'),
    statute miles, ('sm') or nautical miles ('nm').

    The temperature may be in deg C, F, K or R.

    If the units are not specified, the units in default_units.py are used.

    This first version only works for CAS < 661.48 kt.

    """

    # check to confirm the speed is less than 661.48 kt

    ktas = U.speed_conv(tas, from_units=speed_units, to_units="kt")
    if ktas > 661.48:
        raise ValueError(
            "The function tas2dp only works if the tas is less than or equal to 661.48 kt"
        )

    P = SA.alt2press(altitude, alt_units=alt_units, press_units="pa")

    press_ratio = SA.alt2press_ratio(altitude, alt_units)
    temp_ratio = U.temp_conv(temp, from_units=temp_units, to_units="K") / 288.15
    density_ratio = press_ratio / temp_ratio
    Rho = Rho0 * density_ratio
    dp = _speed2dp(tas, P, Rho, press_units=press_units, speed_units=speed_units)

    return dp


def cas2eas(
    cas, altitude, speed_units=default_speed_units, alt_units=default_alt_units,
):
    """
    Return the EAS for a given CAS, pressure altitude and temperature.

    The speed units may be 'kt', 'mph', 'km/h', 'm/s' and 'ft/s'.

    The altitude may be in feet ('ft'), metres ('m'), kilometres ('km'),
    statute miles, ('sm') or nautical miles ('nm').

    If the units are not specified, the units in default_units.py are used.
    """

    dp = cas2dp(cas, speed_units)
    eas = dp2eas(dp, altitude, alt_units=alt_units, speed_units=speed_units)

    return eas


def i_cas2eas(data_items):
    """
    Return the EAS for a given CAS, pressure altitude and temp, with
    interactive input from the user.
    """

    # version that goes interactive, if required

    data_items["cas"] = _get_CAS(data_items)
    cas = data_items["cas"]

    data_items["speed_units"] = _get_speed_units(data_items)
    speed_units = data_items["speed_units"]

    data_items["altitude"] = _get_alt(data_items)
    altitude = data_items["altitude"]

    data_items["alt_units"] = _get_alt_units(data_items)
    alt_units = data_items["alt_units"]

    # print()
    # print("CAS = ", cas, speed_units)
    # print("Altitude = ", altitude, alt_units)
    # print()

    eas = cas2eas(cas, altitude, speed_units, alt_units)
    data_items["eas"] = eas
    return_string = "EAS = " + str(eas) + " " + speed_units
    # print(return_string)


def cas2tas(
    cas,
    altitude,
    temp="std",
    speed_units=default_speed_units,
    alt_units=default_alt_units,
    temp_units=default_temp_units,
):
    """
    Return the TAS for a given CAS, pressure altitude and temperature.

    The speed units may be 'kt', 'mph', 'km/h', 'm/s' and 'ft/s'.

    The altitude may be in feet ('ft'), metres ('m'), kilometres ('km'),
    statute miles, ('sm') or nautical miles ('nm').

    The temperature may be in deg C, F, K or R. The temperature defaults to std
    temperature if it is not input.

    If the units are not specified, the units in default_units.py are used.

    """

    if temp == "std":
        temp = SA.alt2temp(altitude, temp_units=temp_units, alt_units=alt_units)

    dp = cas2dp(cas, speed_units)
    tas = dp2tas(
        dp,
        altitude,
        temp,
        speed_units=speed_units,
        alt_units=alt_units,
        temp_units=temp_units,
    )

    return tas


def i_cas2tas(data_items):
    """
    Return the TAS for a given CAS, pressure altitude and temp, with
    interactive input from the user.
    """

    # version that goes interactive, if required

    data_items["cas"] = _get_CAS(data_items)
    cas = data_items["cas"]

    data_items["speed_units"] = _get_speed_units(data_items)
    speed_units = data_items["speed_units"]

    data_items["altitude"] = _get_alt(data_items)
    altitude = data_items["altitude"]

    data_items["alt_units"] = _get_alt_units(data_items)
    alt_units = data_items["alt_units"]

    data_items["temp_units"] = _get_temp_units(data_items)
    temp_units = data_items["temp_units"]

    data_items["temp"] = _get_temp(data_items)
    temp = data_items["temp"]

    print()
    print("CAS = ", cas, speed_units)
    print("Altitude = ", altitude, alt_units)
    print("Temperature = ", temp, "deg", temp_units)
    print()

    tas = cas2tas(cas, altitude, temp, speed_units, alt_units, temp_units,)
    data_items["tas"] = tas
    return_string = "TAS = " + str(tas) + " " + speed_units
    print(return_string)


def eas2tas(
    eas,
    altitude,
    temp="std",
    speed_units=default_speed_units,
    alt_units=default_alt_units,
    temp_units=default_temp_units,
):
    """
    Return the TAS for a given EAS, pressure altitude and temperature.

    The speed units may be 'kt', 'mph', 'km/h', 'm/s' and 'ft/s'.

    The altitude may be in feet ('ft'), metres ('m'), kilometres ('km'),
    statute miles, ('sm') or nautical miles ('nm').

    The temperature may be in deg C, F, K or R. The temperature defaults to std
    temperature if it is not input.

    If the units are not specified, the units in default_units.py are used.

    """

    if temp == "std":
        temp = SA.alt2temp(altitude, temp_units=temp_units, alt_units=alt_units)

    dp = eas2dp(eas, altitude, speed_units, alt_units)
    tas = dp2tas(
        dp,
        altitude,
        temp,
        speed_units=speed_units,
        alt_units=alt_units,
        temp_units=temp_units,
    )

    return tas


def i_eas2tas(data_items):
    """
    Return the TAS for a given EAS, pressure altitude and temp, with
    interactive input from the user.
    """

    # version that goes interactive, if required

    data_items["eas"] = _get_EAS(data_items)
    eas = data_items["eas"]

    data_items["speed_units"] = _get_speed_units(data_items)
    speed_units = data_items["speed_units"]

    data_items["altitude"] = _get_alt(data_items)
    altitude = data_items["altitude"]

    data_items["alt_units"] = _get_alt_units(data_items)
    alt_units = data_items["alt_units"]

    data_items["temp_units"] = _get_temp_units(data_items)
    temp_units = data_items["temp_units"]

    data_items["temp"] = _get_temp(data_items)
    temp = data_items["temp"]

    print()
    print("EAS = ", eas, speed_units)
    print("Altitude = ", altitude, alt_units)
    print("Temperature = ", temp, "deg", temp_units)
    print()

    tas = eas2tas(eas, altitude, temp, speed_units, alt_units, temp_units,)
    data_items["tas"] = tas
    return_string = "TAS = " + str(tas) + " " + speed_units
    print(return_string)


def eas2cas(
    eas, altitude, speed_units=default_speed_units, alt_units=default_alt_units,
):
    """
    Return the CAS for a given EAS, pressure altitude and temperature.

    The speed units may be 'kt', 'mph', 'km/h', 'm/s' and 'ft/s'.

    The altitude may be in feet ('ft'), metres ('m'), kilometres ('km'),
    statute miles, ('sm') or nautical miles ('nm').

    If the units are not specified, the default units in default_units.py are used.

    Examples:

    Determine equivalent Air Speed for 250 kt CAS at 10,000 ft:
    >>> cas2eas(250, 10000)
    248.09577137102258

    Determine equivalent Air Speed for 250 mph CAS at 10,000 ft:
    >>> cas2eas(250, 10000, speed_units = 'mph')
    248.54048288757579
    """

    dp = eas2dp(eas, altitude, speed_units, alt_units)
    cas = dp2cas(dp, speed_units=speed_units)

    return cas


def i_eas2cas(data_items):
    """
    Return the CAS for a given EAS, pressure altitude, with
    interactive input from the user.

    The speed units may be 'kt', 'mph', 'km/h', 'm/s' and 'ft/s'.

    The altitude may be in feet ('ft'), metres ('m'), kilometres ('km'),
    statute miles, ('sm') or nautical miles ('nm').

    If the units are not specified, the units in default_units.py are used.

    """

    data_items["eas"] = _get_EAS(data_items)
    eas = data_items["eas"]

    data_items["speed_units"] = _get_speed_units(data_items)
    speed_units = data_items["speed_units"]

    data_items["altitude"] = _get_alt(data_items)
    altitude = data_items["altitude"]

    data_items["alt_units"] = _get_alt_units(data_items)
    alt_units = data_items["alt_units"]

    print()
    print("EAS = ", eas, speed_units)
    print("Altitude = ", altitude, alt_units)
    print()

    cas = eas2cas(eas, altitude, speed_units, alt_units)
    data_items["cas"] = cas
    return_string = "CAS = " + str(cas) + " " + speed_units
    print(return_string)


def tas2cas(
    tas,
    altitude,
    temp="std",
    speed_units=default_speed_units,
    alt_units=default_alt_units,
    temp_units=default_temp_units,
):
    """
    Return the CAS for a given TAS, pressure altitude and temperature.

    The speed units may be 'kt', 'mph', 'km/h', 'm/s' and 'ft/s'.

    The altitude may be in feet ('ft'), metres ('m'), kilometres ('km'),
    statute miles, ('sm') or nautical miles ('nm').

    The temperature may be in deg C, F, K or R. The temperature defaults to std
    temperature if it is not input.

    If the units are not specified, the units in default_units.py are used.

    Examples:

    Determine the true Air Speed for 250 kt CAS at 10,000 ft with standard
    temperature:
    >>> cas2tas(250, 10000)
    288.70227231079861

    Determine the true Air Speed for 250 mph CAS at 10,000 ft with standard
    temperature:
    >>> cas2tas(250, 10000, speed_units = 'mph')
    289.21977095514148

    Determine the true Air Speed for 250 mph CAS at 10,000 ft with
    temperature of 0 deg C:
    >>> cas2tas(250, 10000, 0, speed_units = 'mph')
    291.80148048806217

    Determine the true Air Speed for 250 mph CAS at 10,000 ft with
    temperature of 0 deg F:
    >>> cas2tas(250, 10000, 0, speed_units = 'mph', temp_units = 'F')
    282.14588227473797
    """

    if temp == "std":
        temp = SA.alt2temp(altitude, temp_units=temp_units, alt_units=alt_units)

    dp = tas2dp(
        tas, altitude, temp, speed_units, alt_units=alt_units, temp_units=temp_units,
    )
    cas = dp2cas(dp, speed_units=speed_units)

    return cas


def i_tas2cas(data_items):
    """
    Return the CAS for a given TAS, pressure altitude and temp, with
    interactive input from the user.

    The speed units may be 'kt', 'mph', 'km/h', 'm/s' and 'ft/s'.

    The altitude may be in feet ('ft'), metres ('m'), kilometres ('km'),
    statute miles, ('sm') or nautical miles ('nm').

    The temperature may be in deg C, F, K or R. The temperature defaults to std
    temperature if it is not input.

    If the units are not specified, the units in default_units.py are used.

    """

    data_items["tas"] = _get_TAS(data_items)
    tas = data_items["tas"]

    data_items["speed_units"] = _get_speed_units(data_items)
    speed_units = data_items["speed_units"]

    data_items["altitude"] = _get_alt(data_items)
    altitude = data_items["altitude"]

    data_items["alt_units"] = _get_alt_units(data_items)
    alt_units = data_items["alt_units"]

    data_items["temp_units"] = _get_temp_units(data_items)
    temp_units = data_items["temp_units"]

    data_items["temp"] = _get_temp(data_items)
    temp = data_items["temp"]

    print()
    print("TAS = ", tas, speed_units)
    print("Altitude = ", altitude, alt_units)
    print("Temperature = ", temp, "deg", temp_units)
    print()

    cas = tas2cas(tas, altitude, temp, speed_units, alt_units, temp_units,)
    data_items["cas"] = cas
    return_string = "CAS = " + str(cas) + " " + speed_units
    print(return_string)


def tas2eas(
    tas,
    altitude,
    temp="std",
    speed_units=default_speed_units,
    alt_units=default_alt_units,
    temp_units=default_temp_units,
):
    """
    Return the EAS for a given TAS, pressure altitude and temperature.

    The speed units may be 'kt', 'mph', 'km/h', 'm/s' and 'ft/s'.

    The altitude may be in feet ('ft'), metres ('m'), kilometres ('km'),
    statute miles, ('sm') or nautical miles ('nm').

    The temperature may be in deg C, F, K or R. The temperature defaults to std
    temperature if it is not input.

    If the units are not specified, the units in default_units.py are used.

    """

    if temp == "std":
        temp = SA.alt2temp(altitude, temp_units=temp_units, alt_units=alt_units)

    dp = tas2dp(
        tas, altitude, temp, speed_units, alt_units=alt_units, temp_units=temp_units,
    )
    eas = dp2eas(dp, altitude, alt_units=alt_units, speed_units=speed_units)

    return eas


def i_tas2eas(data_items):
    """
    Return the EAS for a given TAS, pressure altitude and temp, with
    interactive input from the user.

    The speed units may be 'kt', 'mph', 'km/h', 'm/s' and 'ft/s'.

    The altitude may be in feet ('ft'), metres ('m'), kilometres ('km'),
    statute miles, ('sm') or nautical miles ('nm').

    The temperature may be in deg C, F, K or R. The temperature defaults to std
    temperature if it is not input.

    If the units are not specified, the units in default_units.py are used.

    """

    data_items["tas"] = _get_TAS(data_items)
    tas = data_items["tas"]

    data_items["speed_units"] = _get_speed_units(data_items)
    speed_units = data_items["speed_units"]

    data_items["altitude"] = _get_alt(data_items)
    altitude = data_items["altitude"]

    data_items["alt_units"] = _get_alt_units(data_items)
    alt_units = data_items["alt_units"]

    data_items["temp_units"] = _get_temp_units(data_items)
    temp_units = data_items["temp_units"]

    data_items["temp"] = _get_temp(data_items)
    temp = data_items["temp"]

    print()
    print("TAS = ", tas, speed_units)
    print("Altitude = ", altitude, alt_units)
    print("Temperature = ", temp, "deg", temp_units)
    print()

    eas = tas2eas(tas, altitude, temp, speed_units, alt_units, temp_units,)
    data_items["eas"] = eas
    return_string = "EAS = " + str(eas) + " " + speed_units
    print(return_string)


# #############################################################################
#
# delta p over p to Mach
#
#     and
#
# Mach to delta p over p
#
# #############################################################################


def dp_over_p2mach(dp_over_p):
    """
    Return the mach number for a given delta p over p.

    Mach must be less than or equal to 10.
    """

    #   mach = (5*( (dp_over_p + 1)**(2/7.) -1) )**0.5

    mach = M.sqrt(5.0 * ((dp_over_p + 1.0) ** (2.0 / 7.0) - 1.0))

    if mach <= 1.0:
        return mach
    else:

        # supersonic case - need to iterate a solution.  Set upper and lower
        # guesses, and iterate until we zero in on a mach that produces the
        # desired result.

        dp_over_p_seek = dp_over_p

        low = 1.0  # initial lower guess for mach

        # This function works up to Mach 10  The upper limit can be
        # extended by increasing the value of the initial upper guess
        # ("high").

        high = 10  # initial upper guess for mach

        # confirm initial low and high are OK:

        dp_over_p_low = mach2dp_over_p(low)
        if dp_over_p_low > dp_over_p_seek:
            raise ValueError("Initial lower mach guess is too high.")

        dp_over_p_high = mach2dp_over_p(high)
        if dp_over_p_high < dp_over_p_seek:
            raise ValueError("Initial upper mach guess is too low.")

        guess = (low + high) / 2.0
        dp_over_p_guess = mach2dp_over_p(guess)

        # keep iterating until dp is within 0.001% of desired value

        while M.fabs(dp_over_p_guess - dp_over_p_seek) / dp_over_p_seek > 1e-5:
            if dp_over_p_guess > dp_over_p_seek:
                high = guess
            else:
                low = guess

            guess = (low + high) / 2.0
            dp_over_p_guess = mach2dp_over_p(guess)

    return guess


def mach2dp_over_p(M):
    """
    Return the delta p over p for a given mach number.
    The result is equal to:
    (pitot pressure - static pressure) / static pressure

    Example - determine the delta p over p at mach 0.4:

    >>> mach2dp_over_p(.4)
    0.11655196580975336
    """

    if M <= 1.0:
        dp_over_p = (M ** 2.0 / 5.0 + 1.0) ** 3.5 - 1.0
    else:
        dp_over_p = (F * M ** 7.0) / (7.0 * M ** 2.0 - 1.0) ** 2.5 - 1.0

    return dp_over_p


# #############################################################################
#
# conversions between cas, mach and altitude
#
# pick any two values, and find the third
#
# #############################################################################


def cas_mach2alt(
    cas, mach, speed_units=default_speed_units, alt_units=default_alt_units,
):
    """
    Return the altitude that corresponds to a given CAS and mach.

    The speed units may be 'kt', 'mph', 'km/h', 'm/s' and 'ft/s'.

    The altitude may be in feet ('ft'), metres ('m'), kilometres ('km'),
    statute miles, ('sm') or nautical miles ('nm').

    If the units are not specified, the units in default_units.py are used.

    """

    dp = cas2dp(cas, speed_units=speed_units, press_units="pa")
    dp_over_p = mach2dp_over_p(mach)
    p = dp / dp_over_p
    altitude = SA.press2alt(p, press_units="pa", alt_units=alt_units)

    return altitude


def i_cas_mach2alt(data_items):
    """
    Return the altitude that corresponds to a given CAS and mach, with an
    interactive interface.
    """

    data_items["cas"] = _get_CAS(data_items)
    cas = data_items["cas"]

    data_items["speed_units"] = _get_speed_units(data_items)
    speed_units = data_items["speed_units"]

    data_items["mach"] = _get_mach(data_items)
    mach = data_items["mach"]

    data_items["alt_units"] = _get_alt_units(data_items)
    alt_units = data_items["alt_units"]

    print()
    print("CAS = ", cas, speed_units)
    print("Mach = ", mach)

    #   print 'Desired altitude units are: ', alt_units

    print()

    alt = cas_mach2alt(cas, mach, speed_units, alt_units)
    data_items["altitude"] = alt

    return_string = "Altitude = " + str(alt) + " " + alt_units
    print(return_string)


def cas_alt2mach(
    cas, altitude, speed_units=default_speed_units, alt_units=default_alt_units,
):
    """
    Return the mach that corresponds to a given CAS and altitude.

    The speed units may be 'kt', 'mph', 'km/h', 'm/s' and 'ft/s'.

    The altitude may be in feet ('ft'), metres ('m'), kilometres ('km'),
    statute miles, ('sm') or nautical miles ('nm').

    If the units are not specified, the units in default_units.py are used.

    """

    dp = cas2dp(cas, speed_units=speed_units, press_units="pa")
    p = SA.alt2press(altitude, alt_units=alt_units, press_units="pa")
    dp_over_p = dp / p
    mach = dp_over_p2mach(dp_over_p)

    return mach


def i_cas_alt2mach(data_items):
    """
    Return the mach that corresponds to a given CAS and altitude, using an
    interactive interface.
    """

    data_items["cas"] = _get_CAS(data_items)
    cas = data_items["cas"]

    data_items["speed_units"] = _get_speed_units(data_items)
    speed_units = data_items["speed_units"]

    data_items["altitude"] = _get_alt(data_items)
    altitude = data_items["altitude"]

    data_items["alt_units"] = _get_alt_units(data_items)
    alt_units = data_items["alt_units"]

    print()
    print("CAS = ", cas, speed_units)
    print("Altitude = ", altitude, alt_units)
    print()

    mach = cas_alt2mach(cas, altitude, speed_units, alt_units)
    data_items["mach"] = mach
    print("Mach = ", mach)


def _cas_alt2mach2(
    cas, altitude, speed_units=default_speed_units, alt_units=default_alt_units,
):
    """
    Alternative, trial variant of cas_alt2mach, using the equations from
    USAF TPS notes.

    The speed units may be 'kt', 'mph', 'km/h', 'm/s' and 'ft/s'.

    The altitude may be in feet ('ft'), metres ('m'), kilometres ('km'),
    statute miles, ('sm') or nautical miles ('nm').

    If the units are not specified, the units in default_units.py are used.

    """

    PR = SA.alt2press_ratio(altitude, alt_units)
    cas = U.speed_conv(cas, from_units=speed_units, to_units="m/s")

    if cas <= A0:

        # <= 661.48 kt

        mach = M.sqrt(
            5.0
            * (
                ((1.0 / PR) * ((1.0 + 0.2 * (cas / A0) ** 2.0) ** 3.5 - 1.0) + 1.0)
                ** (2.0 / 7.0)
                - 1.0
            )
        )
    else:
        raise ValueError("CAS too high.")

    return mach


def mach_alt2cas(
    mach, altitude, alt_units=default_alt_units, speed_units=default_speed_units,
):
    """
    Return the calibrated Air Speed that corresponds to a given mach and
    altitude.

    The speed units may be 'kt', 'mph', 'km/h', 'm/s' and 'ft/s'.

    The altitude may be in feet ('ft'), metres ('m'), kilometres ('km'),
    statute miles, ('sm') or nautical miles ('nm').

    If the units are not specified, the units in default_units.py are used.

    """

    p = SA.alt2press(altitude, alt_units=alt_units, press_units="pa")
    dp_over_p = mach2dp_over_p(mach)
    dp = dp_over_p * p
    cas = dp2cas(dp, press_units="pa", speed_units=speed_units)

    return cas


def i_mach_alt2cas(data_items):
    """
    Return the calibrated Air Speed that corresponds to a given mach and
    altitude, using an interactive interface.
    """

    data_items["mach"] = _get_mach(data_items)
    mach = data_items["mach"]

    data_items["altitude"] = _get_alt(data_items)
    altitude = data_items["altitude"]

    data_items["alt_units"] = _get_alt_units(data_items)
    alt_units = data_items["alt_units"]

    data_items["speed_units"] = _get_speed_units(data_items)
    speed_units = data_items["speed_units"]

    print()
    print("Altitude = ", altitude, alt_units)
    print("Mach = ", mach)
    print()

    cas = mach_alt2cas(mach, altitude, alt_units, speed_units)
    data_items["cas"] = cas
    return_string = "CAS = " + str(cas) + " " + speed_units
    print(return_string)


# #############################################################################
#
# Mach and temperature to TAS
#
#     and
#
# TAS and temperature to Mach
#
# #############################################################################


def mach2tas(
    mach,
    temp="std",
    altitude="blank",
    temp_units=default_temp_units,
    alt_units=default_alt_units,
    speed_units=default_speed_units,
):
    """
    Return the TAS for a given mach number.  The temperature or altitude
    must also be specified.  If the altitude is specified, the temperature
    is assumed to be standard.  If both the altitude and temperature are
    specified, the altitude input is ignored.

    The speed units may be 'kt', 'mph', 'km/h', 'm/s' and 'ft/s'.

    The altitude may be in feet ('ft'), metres ('m'), kilometres ('km'),
    statute miles, ('sm') or nautical miles ('nm').

    The temperature may be in deg C, F, K or R. The temperature defaults to std
    temperature if it is not input.

    If the units are not specified, the units in default_units.py are used.

    Examples:

    Determine the TAS in kt at 0.8 mach at a temperature of
    -15 deg C:
    >>> mach2tas(0.8, -15)
    500.87884108468597

    Determine the TAS in kt at 0.8 mach at 30,000 ft, assuming
    standard temperature:
    >>> mach2tas(0.8, altitude = 30000)
    471.45798523415107

    Determine the TAS in mph at 0.8 mach at 5000 m, assuming
    standard temperature:
    >>> mach2tas(0.8, altitude = 5000, alt_units = 'm', speed_units = 'mph')
    573.60326790383715

    Determine the TAS in km/h at 0.4 mach at a temperature of
    300 deg K:
    >>> mach2tas(0.4, 300, temp_units = 'K', speed_units = 'km/h')
    499.99796329569176
    """

    if temp == "std":
        if altitude != "blank":
            temp = SA.alt2temp(altitude, temp_units=temp_units, alt_units=alt_units)
        else:
            raise ValueError(
                "At least one of the temperature or altitude must be specified."
            )

    tas = mach * SA.temp2speed_of_sound(temp, temp_units, speed_units)

    return tas


def i_mach2tas(data_items):
    """
    Return the TAS that corresponds to a given Mach, altitude, and temperature
    using an interactive interface.
    """

    data_items["mach"] = _get_mach(data_items)
    mach = data_items["mach"]

    data_items["altitude"] = _get_alt(data_items)
    altitude = data_items["altitude"]

    data_items["alt_units"] = _get_alt_units(data_items)
    alt_units = data_items["alt_units"]

    data_items["temp_units"] = _get_temp_units(data_items)
    temp_units = data_items["temp_units"]

    data_items["temp"] = _get_temp(data_items)
    temp = data_items["temp"]

    data_items["speed_units"] = _get_speed_units(data_items)
    speed_units = data_items["speed_units"]

    print()
    print("Mach = ", mach)
    print("Altitude = ", altitude, alt_units)
    print("Temperature =", temp, temp_units)
    print()

    tas = mach2tas(mach, temp, altitude, temp_units, alt_units, speed_units,)
    data_items["tas"] = tas
    print("TAS = ", tas, speed_units)


def tas2mach(
    tas,
    temp="std",
    altitude="blank",
    temp_units=default_temp_units,
    alt_units=default_alt_units,
    speed_units=default_speed_units,
):
    """
    Return the mach number for a given TAS.  The temperature or altitude
    must also be specified.  If the altitude is specified, the temperature
    is assumed to be standard.  If both the altitude and temperature are
    specified, the altitude input is ignored.

    The speed units may be 'kt', 'mph', 'km/h', 'm/s' and 'ft/s'.

    The altitude may be in feet ('ft'), metres ('m'), kilometres ('km'),
    statute miles, ('sm') or nautical miles ('nm').

    The temperature may be in deg C, F, K or R. The temperature defaults to std
    temperature if it is not input.

    If the units are not specified, the units in default_units.py are used.

    Examples:

    Determine the mach number for a TAS of 500 kt at a temperature of
    -15 deg C:
    >>> tas2mach(500, -15)
    0.79859632148519943

    Determine the mach number for a TAS of 500 kt at a temperature of
    0 deg F:
    >>> tas2mach(500, 0, temp_units = 'F')
    0.80292788758764277

    Determine the mach number for a TAS of 500 kt at an altitude of
    10,000 ft, assuming standard temperature:
    >>> tas2mach(500, altitude = 10000)
    0.78328945665870209

    Determine the mach number for a TAS of 400 mph at an altitude of
    5000 m, assuming standard temperature:
    >>> tas2mach(400, altitude = 5000, speed_units = 'mph', alt_units = 'm')
    0.55787687746166581
    """

    if temp == "std":
        if altitude != "blank":
            temp = SA.alt2temp(altitude, temp_units=temp_units, alt_units=alt_units)
        else:
            raise ValueError(
                "At least one of the temperature or altitude must be specified."
            )

    mach = tas / SA.temp2speed_of_sound(temp, temp_units, speed_units)

    return mach


def i_tas2mach(data_items):
    """
    Return the mach that corresponds to a given TAS, altitude, and temperature
    using an interactive interface.
    """

    data_items["tas"] = _get_TAS(data_items)
    tas = data_items["tas"]

    data_items["speed_units"] = _get_speed_units(data_items)
    speed_units = data_items["speed_units"]

    data_items["altitude"] = _get_alt(data_items)
    altitude = data_items["altitude"]

    data_items["alt_units"] = _get_alt_units(data_items)
    alt_units = data_items["alt_units"]

    data_items["temp_units"] = _get_temp_units(data_items)
    temp_units = data_items["temp_units"]

    data_items["temp"] = _get_temp(data_items)
    temp = data_items["temp"]

    print()
    print("TAS = ", tas, speed_units)
    print("Altitude = ", altitude, alt_units)
    print("Temperature =", temp, temp_units)
    print()

    mach = tas2mach(tas, temp, altitude, temp_units, alt_units, speed_units,)
    data_items["mach"] = mach
    print("Mach = ", mach)


# #############################################################################
#
# Ram temperature rise calculations
#
#     Mach and indicated temperature to ambient temperature
#
#         and
#
#     TAS and indicated temperature to ambient temperature
#
# #############################################################################


def mach2temp(
    mach, indicated_temp, recovery_factor, temp_units=default_temp_units,
):
    """
    Return the ambient temp, given the mach number, indicated
    temperature and the temperature probe's recovery factor.

    The temperature may be in deg C, F, K or R.

    If the units are not specified, the units in default_units.py are used.


    Examples:

    Determine the ambient temperature with an indicated temperature of
    -20 deg C at mach 0.6 with a probe recovery factor of 0.8:

    >>> mach2temp(0.6, -20, 0.8)
    -33.787291981845698

    Determine the ambient temperature with an indicated temperature of
    75 deg F at mach 0.3 with a probe recovery factor of 0.9:

    >>> mach2temp(0.3, 75, 0.9, temp_units = 'F')
    66.476427868529839
    """

    indicated_temp = U.temp_conv(indicated_temp, from_units=temp_units, to_units="K")
    ambient_temp = indicated_temp / (1.0 + (0.2 * recovery_factor) * mach ** 2.0)

    ambient_temp = U.temp_conv(ambient_temp, from_units="K", to_units=temp_units)

    return ambient_temp


def tas2temp(
    tas,
    indicated_temp,
    recovery_factor,
    speed_units=default_speed_units,
    temp_units=default_temp_units,
):
    """
    Return the ambient temp, given the TAS, indicated temperature
    and the temperature probe's recovery factor.

    The speed units may be 'kt', 'mph', 'km/h', 'm/s' and 'ft/s'.

    The temperature may be in deg C, F, K or R. The temperature defaults to std
    temperature if it is not input.

    If the units are not specified, the units in default_units.py are used.

    """

    indicated_temp = U.temp_conv(indicated_temp, from_units=temp_units, to_units="K")
    tas = U.speed_conv(tas, from_units=speed_units, to_units="kt")

    # value 7592.4732909142658 was adjusted to make the result equal that
    # obtained using mach2temp

    ambient_temp = indicated_temp - (recovery_factor * tas ** 2.0) / 7592.4732909142658

    ambient_temp = U.temp_conv(ambient_temp, from_units="K", to_units=temp_units)

    return ambient_temp


# #############################################################################
#
# Interactive mode
#
#     Private functions for interactive user interface
#
# #############################################################################


def _get_alt(data_items):  # pragma: no cover
    try:
        prompt = "Altitude = [" + str(data_items["altitude"]) + "] "
        value = VI.get_input2(
            prompt,
            conditions_any=[
                "ERROR - Altitude must be a number.  Commas are not allowed.",
                "type(float(X)) == float",
                'X == ""',
            ],
        )
        if value != "":
            data_items["altitude"] = float(value)
    except KeyError:

        prompt = "Altitude = "
        data_items["altitude"] = float(
            VI.get_input2(
                "Altitude = ",
                conditions_any=[
                    "ERROR - Altitude must be a number.  Commas are not allowed.",
                    "type(float(X)) == float",
                    'X == ""',
                ],
            )
        )
    return data_items["altitude"]


def _get_alt_units(data_items):  # pragma: no cover

    if data_items["alt_units"] == "ft":
        prompt = "altitude units = [ft], m, km, sm, nm: "
    elif data_items["alt_units"] == "m":
        prompt = "altitude units = [m], ft, km, sm, nm: "
    elif data_items["alt_units"] == "km":
        prompt = "altitude units = [km], ft, m, sm, nm: "
    elif data_items["alt_units"] == "sm":
        prompt = "altitude units = [sm], ft, m, km, nm: "
    elif data_items["alt_units"] == "nm":
        prompt = "altitude units = [nm], ft, m, km, sm: "

    alt_units = VI.get_input2(
        prompt,
        conditions_any=[
            'ERROR - altitude units must be one of "ft", "m", "km", "sm" or "nm".',
            'X == "ft"',
            'X =="m"',
            'X == "km"',
            'X == "sm"',
            'X == "nm"',
            'X == ""',
        ],
    )
    if alt_units != "":
        print("Alt units not blank")
        data_items["alt_units"] = alt_units
    return data_items["alt_units"]


def _get_CAS(data_items):  # pragma: no cover
    try:
        prompt = "CAS = [" + str(data_items["cas"]) + "] "
        value = VI.get_input2(
            prompt,
            conditions_any=[
                "ERROR - CAS must be a positive number.",
                "float(X) >= 0",
                'X == ""',
            ],
        )
        if value != "":
            data_items["cas"] = float(value)
    except KeyError:
        prompt = "CAS = "
        data_items["cas"] = float(
            VI.get_input2(
                prompt,
                conditions_any=[
                    "ERROR - CAS must be a positive number.",
                    "float(X) >= 0",
                ],
            )
        )

    return data_items["cas"]


def _get_EAS(data_items):  # pragma: no cover
    try:
        prompt = "EAS = [" + str(data_items["eas"]) + "] "
        value = VI.get_input2(
            prompt,
            conditions_any=[
                "ERROR - EAS must be a positive number.",
                "float(X) >= 0",
                'X == ""',
            ],
        )
        if value != "":
            data_items["eas"] = float(value)
    except KeyError:
        prompt = "EAS = "
        data_items["eas"] = float(
            VI.get_input2(
                prompt,
                conditions_any=[
                    "ERROR - EAS must be a positive number.",
                    "float(X) >= 0",
                ],
            )
        )

    return data_items["eas"]


def _get_TAS(data_items):  # pragma: no cover
    try:
        prompt = "TAS = [" + str(data_items["tas"]) + "] "
        value = VI.get_input2(
            prompt,
            conditions_any=[
                "ERROR - TAS must be a positive number.",
                "float(X) >= 0",
                'X == ""',
            ],
        )
        if value != "":
            data_items["tas"] = float(value)
    except KeyError:
        prompt = "TAS = "
        data_items["tas"] = float(
            VI.get_input2(
                prompt,
                conditions_any=[
                    "ERROR - TAS must be a positive number.",
                    "float(X) >= 0",
                ],
            )
        )

    return data_items["tas"]


def _get_speed_units(data_items):  # pragma: no cover
    if data_items["speed_units"] == "kt":
        prompt = "speed units = [kt], mph, km/h, m/s, ft/s: "
    elif data_items["speed_units"] == "mph":
        prompt = "speed units = [mph], kt, km/h, m/s, ft/s: "
    elif data_items["speed_units"] == "km/h":
        prompt = "speed units = [km/h], kt, mph, m/s, ft/s: "
    elif data_items["speed_units"] == "m/s":
        prompt = "speed units = [m/s], kt, mph, km/h, ft/s: "
    elif data_items["speed_units"] == "mph":
        prompt = "speed units = [ft/s], kt, mph, km/h, m/s: "

    spd_units = VI.get_input2(
        prompt,
        conditions_any=[
            'ERROR - speed units must be one of "kt", "mph", "km/h", "m/s" or "ft/s".',
            'X == "kt"',
            'X =="mph"',
            'X == "km/h"',
            'X == "m/s"',
            'X == "ft/s"',
            'X == ""',
        ],
    )
    if spd_units != "":
        data_items["speed_units"] = spd_units
    return data_items["speed_units"]


def _get_mach(data_items):  # pragma: no cover
    try:
        prompt = "Mach = [" + str(data_items["mach"]) + "] "
        value = VI.get_input2(
            prompt,
            conditions_any=[
                "ERROR - Mach must be a positive number.",
                "float(X) >= 0",
                'X == ""',
            ],
        )
        if value != "":
            data_items["mach"] = float(value)
    except KeyError:
        prompt = "Mach = "
        data_items["mach"] = float(
            VI.get_input2(
                prompt,
                conditions_any=[
                    "ERROR - Mach must be a positive number.",
                    "float(X) >= 0",
                ],
            )
        )

    return data_items["mach"]


def _get_temp(data_items):  # pragma: no cover
    try:
        prompt = "Temperature = [" + str(data_items["temp"]) + "] "
        value = VI.get_input2(
            prompt,
            conditions_any=[
                "ERROR - Temperature must be a number, or S for std temperature.",
                'X =="S"',
                'X == "s"',
                'X == ""',
                "type(float(X)) == float",
            ],
        )
        if value.upper() == "S":
            data_items["temp"] = SA.alt2temp(
                data_items["altitude"],
                alt_units=data_items["alt_units"],
                temp_units=data_items["temp_units"],
            )
        elif value != "":
            data_items["temp"] = value
    except KeyError:
        prompt = "Temperature = [std] "
        value = VI.get_input2(
            prompt,
            conditions_any=[
                "ERROR - Temperature must be a number, or S for std temperature.",
                'X =="S"',
                'X == "s"',
                'X == ""',
                "type(float(X)) == float",
            ],
        )
        if value == "S" or value == "s" or value == "":
            data_items["temp"] = SA.alt2temp(
                data_items["altitude"],
                alt_units=data_items["alt_units"],
                temp_units=data_items["temp_units"],
            )
        else:
            data_items["temp"] = value

    return float(data_items["temp"])


def _get_temp_units(data_items):  # pragma: no cover
    if data_items["temp_units"] == "C":
        prompt = "Temperature units = [C], F, K, R: "
    elif data_items["temp_units"] == "F":
        prompt = "Temperature units = [F], C, K, R: "
    elif data_items["temp_units"] == "K":
        prompt = "Temperature units = [K], C, F, R: "
    elif data_items["temp_units"] == "R":
        prompt = "Temperature units = [R], C, F, K: "

    temp_units = VI.get_input2(
        prompt,
        conditions_any=[
            'ERROR - temperature units must be one of "C", "F", "K", or "R".',
            'X.upper() == "C"',
            'X.upper() == "F"',
            'X.upper() == "K"',
            'X.upper() == "R"',
            'X == ""',
        ],
    ).upper()
    if temp_units != "":
        data_items["temp_units"] = temp_units
    return data_items["temp_units"]


def _interactive_mode():  # pragma: no cover
    """
    Provide interactive interface for selected functions.
    """

    data_items = {}
    data_items["speed_units"] = default_speed_units
    data_items["alt_units"] = default_alt_units
    data_items["temp_units"] = default_temp_units

    _interactive_interface(data_items)


def _interactive_interface(data_items):  # pragma: no cover
    """Provide interactive interface to screen.
    """

    func_list = [
        ["i_cas2eas(data_items)", "CAS to EAS"],
        ["i_cas2tas(data_items)", "CAS to TAS"],
        ["i_eas2cas(data_items)", "EAS to CAS"],
        ["i_eas2tas(data_items)", "EAS to TAS"],
        ["i_tas2cas(data_items)", "TAS to CAS"],
        ["i_tas2eas(data_items)", "TAS to EAS"],
        ["i_cas_alt2mach(data_items)", "CAS and Altitude to Mach"],
        ["i_cas_mach2alt(data_items)", "CAS and Mach to Altitude"],
        ["i_mach_alt2cas(data_items)", "Mach and Altitude to CAS"],
        ["i_tas2mach(data_items)", "TAS, Altitude and Temperature to Mach"],
        ["i_mach2tas(data_items)", "Mach, Altitude and Temperature to TAS"],
    ]

    count = 1
    print("The following functions are available:")
    for func in func_list:
        if count < 10:
            print(" ", count, "-", func[1])
        else:
            print("", count, "-", func[1])
        count += 1
    print("  Q - Quit")

    item = VI.get_input2(
        "Select a function to run by number: ",
        conditions_any=[
            "You must enter an integer between 1 and "
            + str(len(func_list))
            + ', or "Q"',
            "0 < int(X) <= " + str(len(func_list)),
            'X == "Q"',
            'X == "q"',
        ],
    )
    if item.upper() == "Q":
        return
    else:
        item = int(item)
    print()
    func_list_num = item - 1
    eval(func_list[func_list_num][0])
    prompt = "\nDo another calculation [Y/n]"
    input_data = input(prompt)
    if input_data == "" or input_data == "Y" or input_data == "y":
        print("\n")
        _interactive_interface(data_items)
    else:
        sys.exit()


if __name__ == "__main__":  # pragma: no cover

    # run doctest to check the validity of the examples in the doc strings.

    import doctest
    import sys

    doctest.testmod(sys.modules[__name__])

    # run the interactive interface

    _interactive_mode()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2008, Kevin Horton
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# *
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The name of Kevin Horton may not be used to endorse or promote products
#       derived from this software without specific prior written permission.
# *
# THIS SOFTWARE IS PROVIDED BY KEVIN HORTON ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL KEVIN HORTON BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# #############################################################################
#
# version 0.10, 17 Mar 2008
#
# Version History:
# vers     date      Notes
# 0.10   17 Mar 08   Initial version.
# #############################################################################
#
# To Do: 1. Add option to use config file in user's home directory to override
#           default_units.py.
#
# #############################################################################

"""
Defines the default units to be used for various modules.
"""

default_area_units = "ft**2"
default_avgas_units = "lb"
default_density_units = "lb/ft**3"
default_length_units = "ft"
default_power_units = "hp"
default_press_units = "in HG"
default_speed_units = "kt"
default_temp_units = "C"
default_weight_units = "lb"
default_vol_units = "ft**3"

default_alt_units = default_length_units



#!/usr/bin/python
# -*- coding: utf-8 -*-

# #############################################################################
# Copyright (c) 2008, Kevin Horton
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# *
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The name of Kevin Horton may not be used to endorse or promote products
#       derived from this software without specific prior written permission.
# *
# THIS SOFTWARE IS PROVIDED BY KEVIN HORTON ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL KEVIN HORTON BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# #############################################################################
#
# version 0.22, 25 Apr 2008
#
# Version History:
# vers     date     Notes
#  0.1   14 May 06  First release. Only has the unit conversions needed
#                   for the std_atm module:
#                   temperature, pressure, length and density.
#
# 0.11              Added speed conversion
#
# 0.20   06 May 07  Reworked to use default units from default_units module
#
# 0.21   24 Mar 08  Added fuel temperature to avgas_conv
#
# 0.22   25 Apr 08  Corrected error in unit validation in press_conv()
# #############################################################################

"""
Convert between various units.
"""

try:
    from .default_units import *
except ImportError:
    default_area_units = "ft**2"
    default_power_units = "hp"
    default_speed_units = "kt"
    default_temp_units = "C"
    default_weight_units = "lb"
    default_press_units = "in HG"
    default_density_units = "lb/ft**3"
    default_length_units = "ft"
    default_alt_units = default_length_units
    default_avgas_units = "lb"
    default_vol_units = "ft**3"


def area_conv(A, from_units=default_area_units, to_units=default_area_units):
    """
    Convert area values between ft**2, in**2, m**2, km**2, sm**2 and nm**2.

    The incoming value is first converted to ft**2, then it is converted to
    desired return value.


    The units default to those specified in default_units.py

    Examples:

    Convert 1 ft**2 to inches**2, with ft**2 already defined as the default
    units:
    >>> area_conv(1, to_units = 'in**2')
    144.0

    Convert 288 square inches to square feet, with ft**2 already defined as the default
    units:
    >>> area_conv(288, from_units = 'in**2')
    2.0

    Convert 10 square metres to square inches:
    >>> area_conv(1000, from_units = 'm**2', to_units = 'in**2')
    1550003.1000061999
    """

    if from_units == "ft**2":
        pass
    elif from_units == "in**2":
        A /= 144.0
    elif from_units == "m**2":
        A /= 0.3048 ** 2
    elif from_units == "km**2":
        A /= 0.0003048 ** 2
    elif from_units == "sm**2":
        A *= 5280.0 ** 2
    elif from_units == "nm**2":
        A *= (1852 / 0.3048) ** 2
    else:
        raise ValueError(
            'from_units must be "ft**2" or "in**2" or "m**2" or "km**2" or "sm**2" (square statute miles) or "nm**2" (square nautical miles).'
        )

    if to_units == "ft**2":
        return A
    elif to_units == "in**2":
        return A * 144.0
    elif to_units == "m**2":
        return A * 0.3048 ** 2
    elif to_units == "km**2":
        return A * 0.0003048 ** 2
    elif to_units == "sm**2":
        return A / 5280.0 ** 2
    elif to_units == "nm**2":
        return A * (0.3048 / 1852) ** 2
    else:
        raise ValueError(
            'from_units must be "ft**2" or "in**2" or "m**2" or "km**2" or "sm**2" (square statute miles) or "nm**2" (square nautical miles).'
        )


def density_conv(D, from_units, to_units):
    """
    Convert density values between kg/m**3, slug/ft**3 and lb/ft**3.

    The incoming value is first converted to kg/m**3, then it is converted
    to desired return value.

    There are no default units. Both the from_units and the to_units must
    be specified.

    Example:

    Convert 1.225 kg per metre cubed to lb per foot cubed:
    >>> density_conv(1.225, from_units = 'kg/m**3', to_units = 'lb/ft**3')
    0.076474253491112101

    """

    if from_units == "kg/m**3":
        pass
    elif from_units == "slug/ft**3":
        D *= 515.37882
    elif from_units == "lb/ft**3":
        D *= 16.018463
    else:
        raise ValueError(
            'from_units must be one of "kg/m**3", "slug/ft**3" and "lb/ft**3".'
        )

    if to_units == "kg/m**3":
        return D
    elif to_units == "slug/ft**3":
        return D / 515.37882
    elif to_units == "lb/ft**3":
        return D / 16.018463
    else:
        raise ValueError(
            'to_units must be one of "kg/m**3", "slug/ft**3" and "lb/ft**3".'
        )


def force_conv(F, from_units=default_weight_units, to_units=default_weight_units):
    """
    Convert force values between lb and N.

    The incoming value is first converted to N, then it is converted to the
    desired return value.
    """

    if from_units == "N":
        pass
    elif from_units == "lb":
        F *= 4.4482216
    else:
        raise ValueError('from_units must be one of "lb" or "N".')

    if to_units == "N":
        pass
    elif to_units == "lb":
        F /= 4.4482216
    else:
        raise ValueError('to_units must be one of "lb" or "N".')

    return F


def len_conv(L, from_units=default_length_units, to_units=default_length_units):
    """
    Convert length values between ft, in, m, km, sm and nm.

    The incoming value is first converted to ft, then it is converted to
    desired return value.

    The units default to those specified in default_units.py

    Examples:

    Convert 5280 ft to statute miles, with feet already defined as the default
    units:
    >>> len_conv(5280, to_units = 'sm')
    1.0

    Convert 1 nautical mile to feet, with feet already defined as the default
    units:
    >>> len_conv(1, from_units = 'nm')
    6076.1154855643044

    Convert 1000 metres to kilometres:
    >>> len_conv(1000, from_units = 'm', to_units = 'km')
    0.99999999999999989
    """

    if from_units == "ft":
        pass
    elif from_units == "m":
        L /= 0.3048
    elif from_units == "km":
        L /= 0.0003048
    elif from_units == "sm":
        L *= 5280.0
    elif from_units == "nm":
        L *= 1852 / 0.3048
    elif from_units == "in":
        L /= 12.0
    else:
        raise ValueError(
            'from_units must be "ft", "in", "m", "km", "sm" (statute miles) or "nm" (nautical miles).'
        )

    if to_units == "ft":
        return L
    elif to_units == "m":
        return L * 0.3048
    elif to_units == "km":
        return L * 0.0003048
    elif to_units == "sm":
        return L / 5280.0
    elif to_units == "nm":
        return (L * 0.3048) / 1852
    elif to_units == "in":
        return L * 12.0
    else:
        raise ValueError(
            'from_units must be "ft", "in", "m", "km", "sm" (statute miles) or "nm" (nautical miles).'
        )


def power_conv(P, from_units=default_power_units, to_units=default_power_units):
    """
    Convert power values between horsepower, ft-lb/mn,  ft-lb/s, watts,
    kilowatts, BTU/hr and BTU/mn.

    The incoming value is first converted to hp, then it is converted to the
    desired return value.
    The units default to those specified in default_units.py

    """

    if from_units == "hp":
        pass
    elif from_units == "ft-lb/mn":
        P /= 33000.0
    elif from_units == "ft-lb/s":
        P /= 550.0
    elif from_units == "W":
        P /= 745.69987
    elif from_units == "kW":
        P /= 0.74569987
    # elif from_units == 'BTU/hr':
    #     P /= 2544.4332
    # elif from_units == 'BTU/mn':
    #     P /= 42.407227
    else:
        raise ValueError(
            'from_units must be "hp", "ft-lb/mn", "ft-lb/s", "W" (watts), "kW" (kilowatts), "BTU/hr", or "BTU/mn".'
        )

    if to_units == "hp":
        return P
    elif to_units == "ft-lb/mn":
        return P * 33000.0
    elif to_units == "ft-lb/s":
        return P * 550.0
    elif to_units == "W":
        return P * 745.69987
    elif to_units == "kW":
        return P * 0.74569987
    # elif to_units == 'BTU/hr':
    #     return P * 2544.4332
    # elif to_units == 'BTU/mn':
    #     return P * 42.407227
    else:
        raise ValueError(
            'to_units must be "hp", "ft-lb/mn", "ft-lb/s", "W" (watts), "kW" (kilowatts), "BTU/hr", or "BTU/mn".'
        )


def press_conv(P, from_units=default_press_units, to_units=default_press_units):
    """
    Convert pressure values between inches of HG, mm of HG, psi, lb/ft^2,
    hpa and mb.

    The incoming value is first converted to Pa, then it is converted to
    desired return value.
    The units default to those specified in default_units.py


    Examples:

    Convert 1013.25 hpa to default pressure units:
    >>> press_conv(1013.25, from_units = 'hpa')
    29.921331923765198

    Convert 29.9213 default pressure units to mm of HG:
    >>> press_conv(29.9213, to_units = 'mm HG')
    760.00128931459176

    Convert 2116.22 lb per sq. ft to lb per sq. inch:
    >>> press_conv(2116.22, from_units = 'psf', to_units = 'psi')
    14.695973160069311
    """

    if from_units == "in HG":
        P *= 3386.38  # from NASA Reference Publication 1046
    elif from_units == "mm HG":
        P *= 133.322  # derived from NASA Reference Publication 1046 value
    elif from_units == "psi":
        P *= 6894.757  # derived from NASA Reference Publication 1046 value
    elif from_units == "psf" or from_units == "lb/ft**2":

        P *= 47.88026  # from NASA Reference Publication 1046
    elif from_units == "hpa" or from_units == "mb":
        P *= 100.0
    elif from_units == "pa":
        pass
    else:
        raise ValueError(
            'from_units must be "in HG", "mm HG", "psi", "psf" (lb per sq. ft), "hpa", "mb" or "pa".'
        )

    if to_units == "in HG":
        return P / 3386.38
    elif to_units == "mm HG":
        return P / 133.322
    elif to_units == "psi":
        return P / 6894.757
    elif to_units == "psf" or to_units == "lb/ft**2":
        return P / 47.88026
    elif to_units == "hpa" or to_units == "mb":
        return P / 100.0
    elif to_units == "pa":
        return P
    else:
        raise ValueError(
            'to_units must be "in HG", "mm HG", "psi", "psf" (lb per sq. ft), "pa", "hpa" or "mb".'
        )


def speed_conv(S, from_units=default_speed_units, to_units=default_speed_units):
    """
    Convert speed values between kt, mph, km/h, m/s and ft/s.

    The incoming value is first converted to kt, then it is converted to
    desired return value.
    The units default to those specified in default_units.py


    Example:

    Convert 230 mph  to kt:
    >>> speed_conv(230, from_units = 'mph', to_units = 'kt')
    199.86453563714903

    """

    if from_units == "kt":
        pass
    elif from_units == "mph":
        S *= len_conv(1, from_units="sm", to_units="nm")
    elif from_units == "km/h":
        S *= len_conv(1, from_units="km", to_units="nm")
    elif from_units == "m/s":
        S *= len_conv(1, from_units="m", to_units="nm") * 3600.0
    elif from_units == "ft/s":
        S *= len_conv(1, from_units=default_length_units, to_units="nm") * 3600.0
    else:
        raise ValueError(
            'from_units must be one of "kt", "mph", "km/h", "m/s" and "ft/s".'
        )

    if to_units == "kt":
        return S
    elif to_units == "mph":
        S *= len_conv(1, from_units="nm", to_units="sm")
        return S
    elif to_units == "km/h":
        S *= len_conv(1, from_units="nm", to_units="km")
        return S
    elif to_units == "m/s":
        S *= len_conv(1, from_units="nm", to_units="m")
        return S / 3600.0
    elif to_units == "ft/s":
        S *= len_conv(1, from_units="nm", to_units=default_length_units)
        return S / 3600.0
    else:
        raise ValueError(
            'to_units must be one of "kt", "mph", "km/h", "m/s" and "ft/s".'
        )


def temp_conv(T, from_units=default_temp_units, to_units=default_temp_units):
    """
    Convert absolute temperature values between deg C, F, K and R.

    This function should not be used for relative temperature conversions,
    i.e. temperature differences.

    The incoming value is first converted to deg K, then it is converted to
    desired return value.
    The units default to those specified in default_units.py


    Examples:

    Convert 32 deg F to deg C, with deg C as the default units:
    >>> temp_conv(32, from_units = 'F')
    0.0

    Convert 100 deg C to deg F, with deg C as the default units:
    >>> temp_conv(100, to_units = 'F')
    212.0

    Convert 59 deg F to deg K
    >>> temp_conv(59, from_units = 'F', to_units = 'K')
    288.14999999999998
    """

    if from_units == "C":
        T += 273.15
    elif from_units == "F":
        T = ((T - 32) * 5.0) / 9.0 + 273.15
    elif from_units == "K":
        pass
    elif from_units == "R":
        T *= 5.0 / 9.0
    else:
        raise ValueError('from_units must be one of "C", "F", "K" or "R".')

    if to_units == "C":
        return T - 273.15
    elif to_units == "F":
        return (T - 273.15) * 1.8 + 32
    elif to_units == "K":
        return T
    elif to_units == "R":
        return T * 1.8
    else:
        raise ValueError('to_units must be one of "C", "F", "K" or "R".')


def vol_conv(V, from_units=default_vol_units, to_units=default_vol_units):
    """
    Convert volume values between USG, ImpGal (Imperial gallons), l (litres), ft**3, in**3, m**3, km**3, sm**3 and nm**3.

    The incoming value is first converted to ft**3, then it is converted to
    desired return value.


    The units default to those specified in default_units.py

    Examples:

    Convert 1 cubic foot to US gallons, with cubic feet already defined as
    the default units:
    >>> vol_conv(1, to_units = 'USG')
    7.4805194804946105

    Convert 1 Imperial gallon to cubic feet, with cubic feet already defined
    as the default units:
    >>> vol_conv(1, from_units = 'ImpGal')
    0.16054365323600001

    Convert 10 US gallon to litres:
    >>> vol_conv(10, from_units = 'USG', to_units = 'l')
    37.854117840125852
    """

    if from_units == "ft**3":
        pass
    elif from_units == "in**3":
        V /= 12.0 ** 3
    elif from_units == "m**3":
        V /= 0.3048 ** 3
    elif from_units == "km**3":
        V /= 0.0003048 ** 3
    elif from_units == "sm**3":
        V *= 5280.0 ** 3
    elif from_units == "nm**3":
        V *= (1852 / 0.3048) ** 3
    elif from_units == "USG":
        V *= 0.133680555556
    elif from_units == "ImpGal":
        V *= 0.160543653236
    elif from_units == "l":
        V /= 3.048 ** 3
    else:
        raise ValueError(
            'from_units must be "ft**3", "in**3", "USG", "ImpGal", "l", "m**3", "km**3", "sm**3" (cubic statute miles) or "nm**3" (cubic nautical miles).'
        )

    if to_units == "ft**3":
        return V
    elif to_units == "in**3":
        return V * 12.0 ** 3
    elif to_units == "m**3":
        return V * 0.3048 ** 3
    elif to_units == "km**3":
        return V * 0.0003048 ** 3
    elif to_units == "sm**3":
        return V / 5280.0 ** 3
    elif to_units == "nm**3":
        return V * (0.3048 / 1852) ** 3
    elif to_units == "USG":
        return V / 0.133680555556
    elif to_units == "ImpGal":
        return V / 0.160543653236
    elif to_units == "l":
        return V * 3.048 ** 3
    else:
        raise ValueError(
            'to_units must be "ft**3", "in**3", "USG", "ImpGal", "l", "m**3", "km**3", "sm**3" (cubic statute miles) or "nm**3" (cubic nautical miles).'
        )


def wt_conv(W, from_units=default_weight_units, to_units=default_weight_units):
    """
    Convert weight values between lb and kg.

    Purists will yell that lb is a unit of weight, and kg is a unit of mass.
    Get over it.

    The incoming value is first converted to kg, then it is converted to the
    desired return value.

    The units default to those specified in default_units.py


    """

    if from_units == "kg":
        pass
    elif from_units == "lb":
        W *= 0.453592
    else:
        raise ValueError('from_units must be one of "lb" or "kg".')

    if to_units == "kg":
        pass
    elif to_units == "lb":
        W *= 2.204622622
    else:
        raise ValueError('to_units must be one of "lb" or "kg".')

    return W


def avgas_conv(
    AG,
    from_units=default_avgas_units,
    to_units=default_avgas_units,
    temp=15,
    temp_units="C",
    grade="nominal",
):
    """
    Convert aviation gasoline between units of lb, US Gallon (USG),
    Imperial Gallon (Imp Gal), litres (l) and kg, assuming nominal
    density for aviation gasoline of 6.01 lb per USG.

    The units default to those specified in default_units.py

    Note: it was difficult to find authoritative values for aviation gasoline
    density.  Conventional wisdom is that aviation gasoline has a density of
    6 lb/USG.  The Canada Flight Supplement provides densities of:
    temp      density     density    density
    (deg C)   (lb/USG)  (lb/ImpGal)  (lb/l)
    -40         6.41       7.68       1.69
    -20         6.26       7.50       1.65
      0         6.12       7.33       1.62
     15         6.01       7.20       1.59
     30         5.90       7.07       1.56

    However, the Canada Flight Supplement does not provide a source for its
    density data.  And, the values for the different volume units are not
    completly consistent, as they don't vary by exactly the correct factor.
    For example, if the density at 15 deg C is 6.01 lb/USG, we would expect
    the density in lb/ImpGal to be 7.22, (given that 1 ImpGal = 1.201 USG)
    yet the Canada Flight Supplement has 7.20.

    The only authoritative source for aviation gasoline density that was
    found on the web was the \"Air BP Handbook of Products\" on the British
    Petroleum (BP) web site:

    <http://www.bp.com/liveassets/bp_internet/aviation/air_bp/STAGING/local_assets/downloads_pdfs/a/air_bp_products_handbook_04004_1.pdf>

    It provides the following density data valid at 15 deg C (the BP document
    only provides density in kg/m**3 - the density in lb/USG were calculated
    by Kevin Horton):
    Avgas    density     density
    Type    (kg/m**3)    (lb/USG)
    80       690          5.76
    100      695          5.80
    100LL    715          5.97

    The available aviation gasoline specifications do not appear to define an
    allowable density range.  They do define allowable ranges for various
    parametres of the distillation process - the density of the final product
    will vary depending on where in the allowable range the refinery is run.
    Thus there will be some variation in density from refinery to refinery.

    This function uses the 15 deg C density values provided by BP, with the
    variation with temperature provided in the Canada Flight Supplement.

    The grade may be specified as \"80\", \"100\" or \"100LL\".  It defaults to
    \"100LL\" if it is not specified.

    The temperature defaults to 15 deg C if it is not specified.
    """

    lb_per_USG_15_nom = (
        6.01  # nominal density at 15 deg C from Canada Flight Supplement
    )
    slope = (
        -0.007256
    )  # change in density per deg C based on data from Canada Flight Supplement

    lb_per_USG = lb_per_USG_15_nom * (
        1
        + (slope * (temp_conv(temp, from_units=temp_units, to_units="C") - 15))
        / lb_per_USG_15_nom
    )  # density corrected for temperature, using nominal density value

    if grade == "nominal":
        grade_density = lb_per_USG_15_nom
    elif grade == "100LL":
        grade_density = 5.967
    elif str(grade) == "100":
        grade_density = 5.801
    elif str(grade) == "80":
        grade_density = 5.7583
    else:
        raise ValueError(
            'grade must be one of "nominal", "80", "100" or "100LL", with a default of "100LL"'
        )

    # Correct the density if the grade is other than nominal.
    # If the grade actually is nominal, we are multiplying by 1 / 1

    lb_per_USG *= grade_density / lb_per_USG_15_nom

    if from_units == "lb":
        pass
    elif from_units == "USG":
        AG *= lb_per_USG
    elif from_units == "ImpGal":
        AG *= vol_conv(lb_per_USG, from_units="ImpGal", to_units="USG")
    elif from_units == "kg":
        AG = wt_conv(AG, from_units="kg")
    elif from_units == "l":
        AG *= vol_conv(lb_per_USG, from_units="l", to_units="USG")
    else:
        raise ValueError(
            'from_units must be one of "lb", "USG", "Imp Gal", "l", or "kg".'
        )

    if to_units == "lb":
        pass
    elif to_units == "USG":
        AG /= lb_per_USG
    elif to_units == "ImpGal":
        AG /= vol_conv(lb_per_USG, from_units="ImpGal", to_units="USG")
    elif to_units == "kg":
        AG = wt_conv(AG, to_units="kg")
    elif to_units == "l":
        AG /= vol_conv(lb_per_USG, from_units="l", to_units="USG")
    else:
        raise ValueError(
            'from_units must be one of "lb", "USG", "Imp Gal", "l", or "kg".'
        )

    return AG


if __name__ == "__main__":

    # run doctest to check the validity of the examples in the doc strings.

    import doctest
    import sys

    doctest.testmod(sys.modules[__name__])



from distutils.core import setup

setup(
    name="aerocalc3",
    version="0.10",
    description="Aeronautical Engineering Calculations",
    long_description="""
                        AeroCalc is a pure python package that performs various aeronautical
                        engineering calculations.  Currently it provides airspeed conversions,
                        standard atmosphere calculations, static source error correction calculations
                        and unit conversions.
                    """,
    author="Geoffrey Nyaga",
    author_email="geoffrey@geoffreynyaga.com",
    url="https://geoffreynyaga.com/aerocalc3/",
    download_url="https://github.com/geoffreynyaga/aerocalc/archive/v_01.tar.gz",
    packages=["aerocalc3"],
    license="BSD",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: Atmospheric Science",
        "Programming Language :: Python :: 3",
    ],
)
